# Copright Gentoo Foundation 2006
# Portage Unit Testing Functionality

import tempfile
import tarfile
import io
from os import urandom

from portage import os
from portage import shutil
from portage.util._compare_files import compare_files
from portage.tests import TestCase
from portage.tests.resolver.ResolverPlayground import ResolverPlayground
from portage.gpkg import gpkg


class test_gpkg_path_case(TestCase):
    def test_gpkg_short_path(self):
        playground = ResolverPlayground(
            user_config={
                "make.conf": ('BINPKG_COMPRESS="none"',),
            }
        )
        tmpdir = tempfile.mkdtemp()

        try:
            settings = playground.settings
            path_name = (
                "aaaabbbb/ccccdddd/eeeeffff/gggghhhh/iiiijjjj/kkkkllll/"
                "mmmmnnnn/oooopppp/qqqqrrrr/sssstttt/"
            )
            orig_full_path = os.path.join(tmpdir, "orig/" + path_name)
            os.makedirs(orig_full_path)
            with open(os.path.join(orig_full_path, "test"), "wb") as test_file:
                test_file.write(urandom(1048576))

            gpkg_file_loc = os.path.join(tmpdir, "test.gpkg.tar")
            test_gpkg = gpkg(settings, "test", gpkg_file_loc)

            check_result = test_gpkg._check_pre_image_files(
                os.path.join(tmpdir, "orig")
            )
            self.assertEqual(check_result, (95, 4, 0, 1048576, 1048576))

            test_gpkg.compress(os.path.join(tmpdir, "orig"), {"meta": "test"})
            with open(gpkg_file_loc, "rb") as container:
                # container
                self.assertEqual(
                    test_gpkg._get_tar_format(container), tarfile.USTAR_FORMAT
                )

            with tarfile.open(gpkg_file_loc, "r") as container:
                metadata = io.BytesIO(container.extractfile("test/metadata.tar").read())
                self.assertEqual(
                    test_gpkg._get_tar_format(metadata), tarfile.USTAR_FORMAT
                )
                metadata.close()

                image = io.BytesIO(container.extractfile("test/image.tar").read())
                self.assertEqual(test_gpkg._get_tar_format(image), tarfile.USTAR_FORMAT)
                image.close()

            test_gpkg.decompress(os.path.join(tmpdir, "test"))
            r = compare_files(
                os.path.join(tmpdir, "orig/" + path_name + "test"),
                os.path.join(tmpdir, "test/" + path_name + "test"),
                skipped_types=("atime", "mtime", "ctime"),
            )
            self.assertEqual(r, ())
        finally:
            shutil.rmtree(tmpdir)
            playground.cleanup()

    def test_gpkg_long_path(self):
        playground = ResolverPlayground(
            user_config={
                "make.conf": ('BINPKG_COMPRESS="none"',),
            }
        )
        tmpdir = tempfile.mkdtemp()

        try:
            settings = playground.settings

            path_name = (
                "aaaabbbb/ccccdddd/eeeeffff/gggghhhh/iiiijjjj/kkkkllll/"
                "mmmmnnnn/oooopppp/qqqqrrrr/sssstttt/uuuuvvvv/wwwwxxxx/"
                "yyyyzzzz/00001111/22223333/44445555/66667777/88889999/"
                "aaaabbbb/ccccdddd/eeeeffff/gggghhhh/iiiijjjj/kkkkllll/"
                "mmmmnnnn/oooopppp/qqqqrrrr/sssstttt/uuuuvvvv/wwwwxxxx/"
                "yyyyzzzz/00001111/22223333/44445555/66667777/88889999/"
            )
            orig_full_path = os.path.join(tmpdir, "orig/" + path_name)
            os.makedirs(orig_full_path)
            with open(os.path.join(orig_full_path, "test"), "wb") as test_file:
                test_file.write(urandom(1048576))

            gpkg_file_loc = os.path.join(tmpdir, "test.gpkg.tar")
            test_gpkg = gpkg(settings, "test", gpkg_file_loc)

            check_result = test_gpkg._check_pre_image_files(
                os.path.join(tmpdir, "orig")
            )
            self.assertEqual(check_result, (329, 4, 0, 1048576, 1048576))

            test_gpkg.compress(os.path.join(tmpdir, "orig"), {"meta": "test"})
            with open(gpkg_file_loc, "rb") as container:
                # container
                self.assertEqual(
                    test_gpkg._get_tar_format(container), tarfile.USTAR_FORMAT
                )

            with tarfile.open(gpkg_file_loc, "r") as container:
                metadata = io.BytesIO(container.extractfile("test/metadata.tar").read())
                self.assertEqual(
                    test_gpkg._get_tar_format(metadata), tarfile.USTAR_FORMAT
                )
                metadata.close()

                image = io.BytesIO(container.extractfile("test/image.tar").read())
                self.assertEqual(test_gpkg._get_tar_format(image), tarfile.GNU_FORMAT)
                image.close()

            test_gpkg.decompress(os.path.join(tmpdir, "test"))
            r = compare_files(
                os.path.join(tmpdir, "orig/" + path_name + "test"),
                os.path.join(tmpdir, "test/" + path_name + "test"),
                skipped_types=("atime", "mtime", "ctime"),
            )
            self.assertEqual(r, ())
        finally:
            shutil.rmtree(tmpdir)
            playground.cleanup()

    def test_gpkg_non_ascii_path(self):
        playground = ResolverPlayground(
            user_config={
                "make.conf": ('BINPKG_COMPRESS="none"',),
            }
        )
        tmpdir = tempfile.mkdtemp()

        try:
            settings = playground.settings

            path_name = "中文测试/日本語テスト/한국어시험/"
            orig_full_path = os.path.join(tmpdir, "orig/" + path_name)
            os.makedirs(orig_full_path)
            with open(os.path.join(orig_full_path, "test"), "wb") as test_file:
                test_file.write(urandom(1048576))

            gpkg_file_loc = os.path.join(tmpdir, "test.gpkg.tar")
            test_gpkg = gpkg(settings, "test", gpkg_file_loc)

            check_result = test_gpkg._check_pre_image_files(
                os.path.join(tmpdir, "orig")
            )
            self.assertEqual(check_result, (53, 4, 0, 1048576, 1048576))

            test_gpkg.compress(os.path.join(tmpdir, "orig"), {"meta": "test"})
            with open(gpkg_file_loc, "rb") as container:
                # container
                self.assertEqual(
                    test_gpkg._get_tar_format(container), tarfile.USTAR_FORMAT
                )

            with tarfile.open(gpkg_file_loc, "r") as container:
                metadata = io.BytesIO(container.extractfile("test/metadata.tar").read())
                self.assertEqual(
                    test_gpkg._get_tar_format(metadata), tarfile.USTAR_FORMAT
                )
                metadata.close()

                image = io.BytesIO(container.extractfile("test/image.tar").read())
                self.assertEqual(test_gpkg._get_tar_format(image), tarfile.USTAR_FORMAT)
                image.close()

            test_gpkg.decompress(os.path.join(tmpdir, "test"))
            r = compare_files(
                os.path.join(tmpdir, "orig/" + path_name + "test"),
                os.path.join(tmpdir, "test/" + path_name + "test"),
                skipped_types=("atime", "mtime", "ctime"),
            )
            self.assertEqual(r, ())
        finally:
            shutil.rmtree(tmpdir)
            playground.cleanup()

    def test_gpkg_symlink_path(self):
        playground = ResolverPlayground(
            user_config={
                "make.conf": ('BINPKG_COMPRESS="none"',),
            }
        )
        tmpdir = tempfile.mkdtemp()

        try:
            settings = playground.settings

            orig_full_path = os.path.join(tmpdir, "orig/")
            os.makedirs(orig_full_path)
            os.symlink(
                "aaaabbbb/ccccdddd/eeeeffff/gggghhhh/iiiijjjj/kkkkllll/"
                "mmmmnnnn/oooopppp/qqqqrrrr/sssstttt/uuuuvvvv/wwwwxxxx/"
                "yyyyzzzz/00001111/22223333/44445555/66667777/88889999/test",
                os.path.join(orig_full_path, "a_long_symlink"),
            )

            gpkg_file_loc = os.path.join(tmpdir, "test.gpkg.tar")
            test_gpkg = gpkg(settings, "test", gpkg_file_loc)

            check_result = test_gpkg._check_pre_image_files(
                os.path.join(tmpdir, "orig")
            )
            self.assertEqual(check_result, (0, 14, 166, 0, 0))

            test_gpkg.compress(os.path.join(tmpdir, "orig"), {"meta": "test"})
            with open(gpkg_file_loc, "rb") as container:
                # container
                self.assertEqual(
                    test_gpkg._get_tar_format(container), tarfile.USTAR_FORMAT
                )

            with tarfile.open(gpkg_file_loc, "r") as container:
                metadata = io.BytesIO(container.extractfile("test/metadata.tar").read())
                self.assertEqual(
                    test_gpkg._get_tar_format(metadata), tarfile.USTAR_FORMAT
                )
                metadata.close()

                image = io.BytesIO(container.extractfile("test/image.tar").read())
                self.assertEqual(test_gpkg._get_tar_format(image), tarfile.GNU_FORMAT)
                image.close()

            test_gpkg.decompress(os.path.join(tmpdir, "test"))
            r = compare_files(
                os.path.join(tmpdir, "orig/", "a_long_symlink"),
                os.path.join(tmpdir, "test/", "a_long_symlink"),
                skipped_types=("atime", "mtime", "ctime"),
            )
            self.assertEqual(r, ())
        finally:
            shutil.rmtree(tmpdir)
            playground.cleanup()

    def test_gpkg_long_hardlink_path(self):
        playground = ResolverPlayground(
            user_config={
                "make.conf": ('BINPKG_COMPRESS="none"',),
            }
        )
        tmpdir = tempfile.mkdtemp()

        try:
            settings = playground.settings

            path_name = (
                "aaaabbbb/ccccdddd/eeeeffff/gggghhhh/iiiijjjj/kkkkllll/"
                "mmmmnnnn/oooopppp/qqqqrrrr/sssstttt/uuuuvvvv/wwwwxxxx/"
            )
            file_name = (
                "test-A-B-C-D-E-F-G-H-I-J-K-L-M-N-O-P-Q-R-S-T-U-V-W-X-Y-Z"
                "A-B-C-D-E-F-G-H-I-J-K-L-M-N-O-P-Q-R-S-T-U-V-W-X-Y-Z"
                "A-B-C-D-E-F-G-H-I-J-K-L-M-N-O-P-Q-R-S-T-U-V-W-X-Y-Z"
            )
            orig_full_path = os.path.join(tmpdir, "orig", path_name)
            os.makedirs(orig_full_path)
            with open(os.path.join(orig_full_path, "test"), "wb") as test_file:
                test_file.write(urandom(1048576))

            os.link(
                os.path.join(orig_full_path, "test"),
                os.path.join(orig_full_path, file_name),
            )

            gpkg_file_loc = os.path.join(tmpdir, "test.gpkg.tar")
            test_gpkg = gpkg(settings, "test", gpkg_file_loc)

            check_result = test_gpkg._check_pre_image_files(
                os.path.join(tmpdir, "orig")
            )
            self.assertEqual(check_result, (113, 158, 272, 1048576, 2097152))

            test_gpkg.compress(os.path.join(tmpdir, "orig"), {"meta": "test"})
            with open(gpkg_file_loc, "rb") as container:
                # container
                self.assertEqual(
                    test_gpkg._get_tar_format(container), tarfile.USTAR_FORMAT
                )

            with tarfile.open(gpkg_file_loc, "r") as container:
                metadata = io.BytesIO(container.extractfile("test/metadata.tar").read())
                self.assertEqual(
                    test_gpkg._get_tar_format(metadata), tarfile.USTAR_FORMAT
                )
                metadata.close()

                image = io.BytesIO(container.extractfile("test/image.tar").read())
                self.assertEqual(test_gpkg._get_tar_format(image), tarfile.GNU_FORMAT)
                image.close()

            test_gpkg.decompress(os.path.join(tmpdir, "test"))
            r = compare_files(
                os.path.join(tmpdir, "orig", path_name, file_name),
                os.path.join(tmpdir, "test", path_name, file_name),
                skipped_types=("atime", "mtime", "ctime"),
            )
            self.assertEqual(r, ())
        finally:
            shutil.rmtree(tmpdir)

    def test_gpkg_long_filename(self):
        playground = ResolverPlayground(
            user_config={
                "make.conf": ('BINPKG_COMPRESS="none"',),
            }
        )
        tmpdir = tempfile.mkdtemp()

        try:
            settings = playground.settings
            path_name = "aaaabbbb/ccccdddd/eeeeffff/gggghhhh/iiiijjjj/kkkkllll/"
            file_name = (
                "test1234567890"
                "A-B-C-D-E-F-G-H-I-J-K-L-M-N-O-P-Q-R-S-T-U-V-W-X-Y-Z"
                "A-B-C-D-E-F-G-H-I-J-K-L-M-N-O-P-Q-R-S-T-U-V-W-X-Y-Z"
                "A-B-C-D-E-F-G-H-I-J-K-L-M-N-O-P-Q-R-S-T-U-V-W-X-Y-Z"
            )

            orig_full_path = os.path.join(tmpdir, "orig/" + path_name)
            os.makedirs(orig_full_path)
            with open(os.path.join(orig_full_path, file_name), "wb") as test_file:
                test_file.write(urandom(1048576))

            gpkg_file_loc = os.path.join(tmpdir, "test.gpkg.tar")
            test_gpkg = gpkg(settings, "test", gpkg_file_loc)

            check_result = test_gpkg._check_pre_image_files(
                os.path.join(tmpdir, "orig")
            )
            self.assertEqual(check_result, (59, 167, 0, 1048576, 1048576))

            test_gpkg.compress(os.path.join(tmpdir, "orig"), {"meta": "test"})
            with open(gpkg_file_loc, "rb") as container:
                # container
                self.assertEqual(
                    test_gpkg._get_tar_format(container), tarfile.USTAR_FORMAT
                )

            with tarfile.open(gpkg_file_loc, "r") as container:
                metadata = io.BytesIO(container.extractfile("test/metadata.tar").read())
                self.assertEqual(
                    test_gpkg._get_tar_format(metadata), tarfile.USTAR_FORMAT
                )
                metadata.close()

                image = io.BytesIO(container.extractfile("test/image.tar").read())
                self.assertEqual(test_gpkg._get_tar_format(image), tarfile.GNU_FORMAT)
                image.close()

            test_gpkg.decompress(os.path.join(tmpdir, "test"))
            r = compare_files(
                os.path.join(tmpdir, "orig", path_name, file_name),
                os.path.join(tmpdir, "test", path_name, file_name),
                skipped_types=("atime", "mtime", "ctime"),
            )
            self.assertEqual(r, ())
        finally:
            shutil.rmtree(tmpdir)
            playground.cleanup()



# Copyright 2007-2021 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import glob
import time

from portage import os
from portage.exception import PortageKeyError
from portage.versions import best, catsplit, vercmp
from portage.dep import Atom, use_reduce
from portage.dep._slot_operator import strip_slots
from portage.localization import _
from portage._sets.base import PackageSet
from portage._sets import SetConfigError, get_boolean
import portage

__all__ = [
    "CategorySet",
    "ChangedDepsSet",
    "DowngradeSet",
    "EverythingSet",
    "OwnerSet",
    "SubslotChangedSet",
    "VariableSet",
]


class EverythingSet(PackageSet):
    _operations = ["merge"]
    description = (
        "Package set which contains SLOT " + "atoms to match all installed packages"
    )
    _filter = None

    def __init__(self, vdbapi, **kwargs):
        super().__init__()
        self._db = vdbapi

    def load(self):
        myatoms = []
        pkg_str = self._db._pkg_str
        cp_list = self._db.cp_list

        for cp in self._db.cp_all():
            for cpv in cp_list(cp):
                # NOTE: Create SLOT atoms even when there is only one
                # SLOT installed, in order to avoid the possibility
                # of unwanted upgrades as reported in bug #338959.
                pkg = pkg_str(cpv, None)
                atom = Atom("{}:{}".format(pkg.cp, pkg.slot))
                if self._filter:
                    if self._filter(atom):
                        myatoms.append(atom)
                else:
                    myatoms.append(atom)

        self._setAtoms(myatoms)

    def singleBuilder(self, options, settings, trees):
        return EverythingSet(trees["vartree"].dbapi)

    singleBuilder = classmethod(singleBuilder)


class OwnerSet(PackageSet):

    _operations = ["merge", "unmerge"]

    description = (
        "Package set which contains all packages " + "that own one or more files."
    )

    def __init__(self, vardb=None, exclude_files=None, files=None):
        super().__init__()
        self._db = vardb
        self._exclude_files = exclude_files
        self._files = files

    def mapPathsToAtoms(self, paths, exclude_paths=None):
        """
        All paths must begin with a slash, and must not include EROOT.
        Supports globs.
        """
        rValue = set()
        vardb = self._db

        eroot = vardb.settings["EROOT"]
        expanded_paths = []
        for p in paths:
            expanded_paths.extend(
                expanded_p[len(eroot) - 1 :]
                for expanded_p in glob.iglob(os.path.join(eroot, p.lstrip(os.sep)))
            )
        paths = expanded_paths

        expanded_exclude_paths = []
        for p in exclude_paths or ():
            expanded_exclude_paths.extend(
                expanded_exc_p[len(eroot) - 1 :]
                for expanded_exc_p in glob.iglob(os.path.join(eroot, p.lstrip(os.sep)))
            )
        exclude_paths = expanded_exclude_paths

        pkg_str = vardb._pkg_str
        if not exclude_paths:
            for link, p in vardb._owners.iter_owners(paths):
                pkg = pkg_str(link.mycpv, None)
                rValue.add("{}:{}".format(pkg.cp, pkg.slot))
        else:
            all_paths = set()
            all_paths.update(paths)
            all_paths.update(exclude_paths)
            exclude_atoms = set()
            for link, p in vardb._owners.iter_owners(all_paths):
                pkg = pkg_str(link.mycpv, None)
                atom = "{}:{}".format(pkg.cp, pkg.slot)
                rValue.add(atom)
                # Returned paths are relative to ROOT and do not have
                # a leading slash.
                if "/" + p in exclude_paths:
                    exclude_atoms.add(atom)
            rValue.difference_update(exclude_atoms)

        return rValue

    def load(self):
        self._setAtoms(
            self.mapPathsToAtoms(self._files, exclude_paths=self._exclude_files)
        )

    def singleBuilder(cls, options, settings, trees):
        if not "files" in options:
            raise SetConfigError(_("no files given"))

        exclude_files = options.get("exclude-files")
        if exclude_files is not None:
            exclude_files = frozenset(portage.util.shlex_split(exclude_files))
        return cls(
            vardb=trees["vartree"].dbapi,
            exclude_files=exclude_files,
            files=frozenset(portage.util.shlex_split(options["files"])),
        )

    singleBuilder = classmethod(singleBuilder)


class VariableSet(EverythingSet):

    _operations = ["merge", "unmerge"]

    description = (
        "Package set which contains all packages "
        + "that match specified values of a specified variable."
    )

    def __init__(
        self, vardb, metadatadb=None, variable=None, includes=None, excludes=None
    ):
        super().__init__(vardb)
        self._metadatadb = metadatadb
        self._variable = variable
        self._includes = includes
        self._excludes = excludes

    def _filter(self, atom):
        ebuild = best(self._metadatadb.match(atom))
        if not ebuild:
            return False
        (values,) = self._metadatadb.aux_get(ebuild, [self._variable])
        values_list = values.split()

        if "DEPEND" in self._variable:
            include_atoms = []
            for include in self._includes:
                include_atoms.append(Atom(include))

            for x in use_reduce(values, token_class=Atom):
                if not isinstance(x, Atom):
                    continue

                for include_atom in include_atoms:
                    if include_atom.match(x):
                        return True

            return False

        if self._includes and not self._includes.intersection(values_list):
            return False

        if self._excludes and self._excludes.intersection(values_list):
            return False

        return True

    def singleBuilder(cls, options, settings, trees):
        variable = options.get("variable")
        if variable is None:
            raise SetConfigError(_("missing required attribute: 'variable'"))

        includes = options.get("includes", "")
        excludes = options.get("excludes", "")

        if not (includes or excludes):
            raise SetConfigError(_("no includes or excludes given"))

        metadatadb = options.get("metadata-source", "vartree")
        if not metadatadb in trees:
            raise SetConfigError(
                _("invalid value '%s' for option metadata-source") % metadatadb
            )

        return cls(
            trees["vartree"].dbapi,
            metadatadb=trees[metadatadb].dbapi,
            excludes=frozenset(excludes.split()),
            includes=frozenset(includes.split()),
            variable=variable,
        )

    singleBuilder = classmethod(singleBuilder)


class SubslotChangedSet(PackageSet):

    _operations = ["merge", "unmerge"]

    description = (
        "Package set which contains all packages "
        + "for which the subslot of the highest visible ebuild is "
        + "different than the currently installed version."
    )

    def __init__(self, portdb=None, vardb=None):
        super().__init__()
        self._portdb = portdb
        self._vardb = vardb

    def load(self):
        atoms = []
        xmatch = self._portdb.xmatch
        xmatch_level = "bestmatch-visible"
        cp_list = self._vardb.cp_list
        for cp in self._vardb.cp_all():
            for pkg in cp_list(cp):
                slot_atom = "{}:{}".format(pkg.cp, pkg.slot)
                ebuild = xmatch(xmatch_level, slot_atom)
                if not ebuild:
                    continue
                if pkg.sub_slot != ebuild.sub_slot:
                    atoms.append(slot_atom)

        self._setAtoms(atoms)

    def singleBuilder(cls, options, settings, trees):
        return cls(portdb=trees["porttree"].dbapi, vardb=trees["vartree"].dbapi)

    singleBuilder = classmethod(singleBuilder)


class DowngradeSet(PackageSet):

    _operations = ["merge", "unmerge"]

    description = (
        "Package set which contains all packages "
        + "for which the highest visible ebuild version is lower than "
        + "the currently installed version."
    )

    def __init__(self, portdb=None, vardb=None):
        super().__init__()
        self._portdb = portdb
        self._vardb = vardb

    def load(self):
        atoms = []
        xmatch = self._portdb.xmatch
        xmatch_level = "bestmatch-visible"
        cp_list = self._vardb.cp_list
        pkg_str = self._vardb._pkg_str
        for cp in self._vardb.cp_all():
            for cpv in cp_list(cp):
                pkg = pkg_str(cpv, None)
                slot_atom = "{}:{}".format(pkg.cp, pkg.slot)
                ebuild = xmatch(xmatch_level, slot_atom)
                if not ebuild:
                    continue
                if vercmp(cpv.version, ebuild.version) > 0:
                    atoms.append(slot_atom)

        self._setAtoms(atoms)

    def singleBuilder(cls, options, settings, trees):
        return cls(portdb=trees["porttree"].dbapi, vardb=trees["vartree"].dbapi)

    singleBuilder = classmethod(singleBuilder)


class UnavailableSet(EverythingSet):

    _operations = ["unmerge"]

    description = (
        "Package set which contains all installed "
        + "packages for which there are no visible ebuilds "
        + "corresponding to the same $CATEGORY/$PN:$SLOT."
    )

    def __init__(self, vardb, metadatadb=None):
        super().__init__(vardb)
        self._metadatadb = metadatadb

    def _filter(self, atom):
        return not self._metadatadb.match(atom)

    def singleBuilder(cls, options, settings, trees):

        metadatadb = options.get("metadata-source", "porttree")
        if not metadatadb in trees:
            raise SetConfigError(
                _("invalid value '%s' for option " "metadata-source") % (metadatadb,)
            )

        return cls(trees["vartree"].dbapi, metadatadb=trees[metadatadb].dbapi)

    singleBuilder = classmethod(singleBuilder)


class UnavailableBinaries(EverythingSet):

    _operations = (
        "merge",
        "unmerge",
    )

    description = (
        "Package set which contains all installed "
        + "packages for which corresponding binary packages "
        + "are not available."
    )

    def __init__(self, vardb, metadatadb=None):
        super().__init__(vardb)
        self._metadatadb = metadatadb

    def _filter(self, atom):
        inst_pkg = self._db.match(atom)
        if not inst_pkg:
            return False
        inst_cpv = inst_pkg[0]
        return not self._metadatadb.cpv_exists(inst_cpv)

    def singleBuilder(cls, options, settings, trees):

        metadatadb = options.get("metadata-source", "bintree")
        if not metadatadb in trees:
            raise SetConfigError(
                _("invalid value '%s' for option " "metadata-source") % (metadatadb,)
            )

        return cls(trees["vartree"].dbapi, metadatadb=trees[metadatadb].dbapi)

    singleBuilder = classmethod(singleBuilder)


class CategorySet(PackageSet):
    _operations = ["merge", "unmerge"]

    def __init__(self, category, dbapi, only_visible=True):
        super().__init__()
        self._db = dbapi
        self._category = category
        self._check = only_visible
        if only_visible:
            s = "visible"
        else:
            s = "all"
        self.description = "Package set containing {} packages of category {}".format(
            s,
            self._category,
        )

    def load(self):
        myatoms = []
        for cp in self._db.cp_all():
            if catsplit(cp)[0] == self._category:
                if (not self._check) or len(self._db.match(cp)) > 0:
                    myatoms.append(cp)
        self._setAtoms(myatoms)

    def _builderGetRepository(cls, options, repositories):
        repository = options.get("repository", "porttree")
        if not repository in repositories:
            raise SetConfigError(_("invalid repository class '%s'") % repository)
        return repository

    _builderGetRepository = classmethod(_builderGetRepository)

    def _builderGetVisible(cls, options):
        return get_boolean(options, "only_visible", True)

    _builderGetVisible = classmethod(_builderGetVisible)

    def singleBuilder(cls, options, settings, trees):
        if not "category" in options:
            raise SetConfigError(_("no category given"))

        category = options["category"]
        if not category in settings.categories:
            raise SetConfigError(_("invalid category name '%s'") % category)

        repository = cls._builderGetRepository(options, trees.keys())
        visible = cls._builderGetVisible(options)

        return CategorySet(
            category, dbapi=trees[repository].dbapi, only_visible=visible
        )

    singleBuilder = classmethod(singleBuilder)

    def multiBuilder(cls, options, settings, trees):
        rValue = {}

        if "categories" in options:
            categories = options["categories"].split()
            invalid = set(categories).difference(settings.categories)
            if invalid:
                raise SetConfigError(
                    _("invalid categories: %s") % ", ".join(list(invalid))
                )
        else:
            categories = settings.categories

        repository = cls._builderGetRepository(options, trees.keys())
        visible = cls._builderGetVisible(options)
        name_pattern = options.get("name_pattern", "$category/*")

        if not "$category" in name_pattern and not "${category}" in name_pattern:
            raise SetConfigError(
                _("name_pattern doesn't include $category placeholder")
            )

        for cat in categories:
            myset = CategorySet(cat, trees[repository].dbapi, only_visible=visible)
            myname = name_pattern.replace("$category", cat)
            myname = myname.replace("${category}", cat)
            rValue[myname] = myset
        return rValue

    multiBuilder = classmethod(multiBuilder)


class AgeSet(EverythingSet):
    _operations = ["merge", "unmerge"]
    _aux_keys = ("BUILD_TIME",)

    def __init__(self, vardb, mode="older", age=7):
        super().__init__(vardb)
        self._mode = mode
        self._age = age

    def _filter(self, atom):

        cpv = self._db.match(atom)[0]
        try:
            (date,) = self._db.aux_get(cpv, self._aux_keys)
            date = int(date)
        except (KeyError, ValueError):
            return bool(self._mode == "older")
        age = (time.time() - date) / (3600 * 24)
        if (self._mode == "older" and age <= self._age) or (
            self._mode == "newer" and age >= self._age
        ):
            return False
        return True

    def singleBuilder(cls, options, settings, trees):
        mode = options.get("mode", "older")
        if str(mode).lower() not in ["newer", "older"]:
            raise SetConfigError(
                _("invalid 'mode' value %s (use either 'newer' or 'older')") % mode
            )
        try:
            age = int(options.get("age", "7"))
        except ValueError as e:
            raise SetConfigError(_("value of option 'age' is not an integer"))
        return AgeSet(vardb=trees["vartree"].dbapi, mode=mode, age=age)

    singleBuilder = classmethod(singleBuilder)


class DateSet(EverythingSet):
    _operations = ["merge", "unmerge"]
    _aux_keys = ("BUILD_TIME",)

    def __init__(self, vardb, date, mode="older"):
        super().__init__(vardb)
        self._mode = mode
        self._date = date

    def _filter(self, atom):

        cpv = self._db.match(atom)[0]
        try:
            (date,) = self._db.aux_get(cpv, self._aux_keys)
            date = int(date)
        except (KeyError, ValueError):
            return bool(self._mode == "older")
        # Make sure inequality is _strict_ to exclude tested package
        if (self._mode == "older" and date < self._date) or (
            self._mode == "newer" and date > self._date
        ):
            return True
        return False

    def singleBuilder(cls, options, settings, trees):
        vardbapi = trees["vartree"].dbapi
        mode = options.get("mode", "older")
        if str(mode).lower() not in ["newer", "older"]:
            raise SetConfigError(
                _("invalid 'mode' value %s (use either 'newer' or 'older')") % mode
            )

        formats = []
        if options.get("package") is not None:
            formats.append("package")
        if options.get("filestamp") is not None:
            formats.append("filestamp")
        if options.get("seconds") is not None:
            formats.append("seconds")
        if options.get("date") is not None:
            formats.append("date")

        if not formats:
            raise SetConfigError(
                _(
                    "none of these options specified: 'package', 'filestamp', 'seconds', 'date'"
                )
            )
        elif len(formats) > 1:
            raise SetConfigError(
                _(
                    "no more than one of these options is allowed: 'package', 'filestamp', 'seconds', 'date'"
                )
            )

        setformat = formats[0]

        if setformat == "package":
            package = options.get("package")
            try:
                cpv = vardbapi.match(package)[0]
                (date,) = vardbapi.aux_get(cpv, ("BUILD_TIME",))
                date = int(date)
            except (KeyError, ValueError):
                raise SetConfigError(
                    _("cannot determine installation date of package %s") % package
                )
        elif setformat == "filestamp":
            filestamp = options.get("filestamp")
            try:
                date = int(os.stat(filestamp).st_mtime)
            except (OSError, ValueError):
                raise SetConfigError(
                    _("cannot determine 'filestamp' of '%s'") % filestamp
                )
        elif setformat == "seconds":
            try:
                date = int(options.get("seconds"))
            except ValueError:
                raise SetConfigError(_("option 'seconds' must be an integer"))
        else:
            dateopt = options.get("date")
            try:
                dateformat = options.get("dateformat", "%x %X")
                date = int(time.mktime(time.strptime(dateopt, dateformat)))
            except ValueError:
                raise SetConfigError(
                    _("'date=%s' does not match 'dateformat=%s'")
                    % (dateopt, dateformat)
                )
        return DateSet(vardb=vardbapi, date=date, mode=mode)

    singleBuilder = classmethod(singleBuilder)


class RebuiltBinaries(EverythingSet):
    _operations = ("merge",)
    _aux_keys = ("BUILD_TIME",)

    def __init__(self, vardb, bindb=None):
        super().__init__(vardb, bindb=bindb)
        self._bindb = bindb

    def _filter(self, atom):
        cpv = self._db.match(atom)[0]
        (inst_build_time,) = self._db.aux_get(cpv, self._aux_keys)
        try:
            (bin_build_time,) = self._bindb.aux_get(cpv, self._aux_keys)
        except KeyError:
            return False
        return bool(bin_build_time and (inst_build_time != bin_build_time))

    def singleBuilder(cls, options, settings, trees):
        return RebuiltBinaries(trees["vartree"].dbapi, bindb=trees["bintree"].dbapi)

    singleBuilder = classmethod(singleBuilder)


class ChangedDepsSet(PackageSet):

    _operations = ["merge", "unmerge"]

    description = (
        "Package set which contains all installed "
        + "packages for which the vdb *DEPEND entries are outdated "
        + "compared to corresponding portdb entries."
    )

    def __init__(self, portdb=None, vardb=None):
        super().__init__()
        self._portdb = portdb
        self._vardb = vardb

    def load(self):
        depvars = ("RDEPEND", "PDEPEND")
        ebuild_vars = depvars + ("EAPI",)
        installed_vars = depvars + ("USE", "EAPI")

        atoms = []
        for cpv in self._vardb.cpv_all():
            # no ebuild, no update :).
            try:
                ebuild_metadata = dict(
                    zip(ebuild_vars, self._portdb.aux_get(cpv, ebuild_vars))
                )
            except PortageKeyError:
                continue

            # USE flags used to build the ebuild and EAPI
            # (needed for Atom & use_reduce())
            installed_metadata = dict(
                zip(installed_vars, self._vardb.aux_get(cpv, installed_vars))
            )
            usel = frozenset(installed_metadata["USE"].split())

            # get all *DEPEND variables from vdb & portdb and compare them.
            # we need to do some cleaning up & expansion to make matching
            # meaningful since vdb dependencies are conditional-free.
            vdbvars = [
                strip_slots(
                    use_reduce(
                        installed_metadata[k],
                        uselist=usel,
                        eapi=installed_metadata["EAPI"],
                        token_class=Atom,
                    )
                )
                for k in depvars
            ]
            pdbvars = [
                strip_slots(
                    use_reduce(
                        ebuild_metadata[k],
                        uselist=usel,
                        eapi=ebuild_metadata["EAPI"],
                        token_class=Atom,
                    )
                )
                for k in depvars
            ]

            # if dependencies don't match, trigger the rebuild.
            if vdbvars != pdbvars:
                atoms.append("=%s" % cpv)

        self._setAtoms(atoms)

    def singleBuilder(cls, options, settings, trees):
        return cls(portdb=trees["porttree"].dbapi, vardb=trees["vartree"].dbapi)

    singleBuilder = classmethod(singleBuilder)



# Copyright 2006-2020 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

from portage.tests import TestCase
from portage.dep import Atom, match_from_list, _repo_separator
from portage.versions import catpkgsplit, _pkg_str


class Package:
    """
    Provides a minimal subset of attributes of _emerge.Package.Package
    """

    def __init__(self, atom):
        atom = Atom(atom, allow_repo=True)
        self.cp = atom.cp
        slot = atom.slot
        if atom.sub_slot:
            slot = "{}/{}".format(slot, atom.sub_slot)
        if not slot:
            slot = "0"
        self.cpv = _pkg_str(atom.cpv, slot=slot, repo=atom.repo)
        self.cpv_split = catpkgsplit(self.cpv)
        self.slot = self.cpv.slot
        self.sub_slot = self.cpv.sub_slot
        self.repo = atom.repo
        if atom.use:
            self.use = self._use_class(atom.use.enabled)
            self.iuse = self._iuse_class(atom.use.required)
        else:
            self.use = self._use_class([])
            self.iuse = self._iuse_class([])

    class _use_class:
        def __init__(self, use):
            self.enabled = frozenset(use)

    class _iuse_class:
        def __init__(self, iuse):
            self.all = frozenset(iuse)

        def is_valid_flag(self, flags):
            if isinstance(flags, str):
                flags = [flags]
            for flag in flags:
                if not flag in self.all:
                    return False
            return True


class Test_match_from_list(TestCase):
    def testMatch_from_list(self):
        tests = (
            ("=sys-apps/portage-45*", [], []),
            (
                "=sys-apps/portage-45*",
                ["sys-apps/portage-045"],
                ["sys-apps/portage-045"],
            ),
            (
                "!=sys-apps/portage-45*",
                ["sys-apps/portage-045"],
                ["sys-apps/portage-045"],
            ),
            (
                "!!=sys-apps/portage-45*",
                ["sys-apps/portage-045"],
                ["sys-apps/portage-045"],
            ),
            (
                "=sys-apps/portage-045",
                ["sys-apps/portage-045"],
                ["sys-apps/portage-045"],
            ),
            ("=sys-apps/portage-045", ["sys-apps/portage-046"], []),
            (
                "~sys-apps/portage-045",
                ["sys-apps/portage-045-r1"],
                ["sys-apps/portage-045-r1"],
            ),
            ("~sys-apps/portage-045", ["sys-apps/portage-046-r1"], []),
            (
                "<=sys-apps/portage-045",
                ["sys-apps/portage-045"],
                ["sys-apps/portage-045"],
            ),
            ("<=sys-apps/portage-045", ["sys-apps/portage-046"], []),
            (
                "<sys-apps/portage-046",
                ["sys-apps/portage-045"],
                ["sys-apps/portage-045"],
            ),
            ("<sys-apps/portage-046", ["sys-apps/portage-046"], []),
            (
                ">=sys-apps/portage-045",
                ["sys-apps/portage-045"],
                ["sys-apps/portage-045"],
            ),
            (">=sys-apps/portage-047", ["sys-apps/portage-046-r1"], []),
            (
                ">sys-apps/portage-044",
                ["sys-apps/portage-045"],
                ["sys-apps/portage-045"],
            ),
            (">sys-apps/portage-047", ["sys-apps/portage-046-r1"], []),
            (
                "sys-apps/portage:0",
                [Package("=sys-apps/portage-045:0")],
                ["sys-apps/portage-045"],
            ),
            ("sys-apps/portage:0", [Package("=sys-apps/portage-045:1")], []),
            ("=cat/pkg-1-r1*", ["cat/pkg-1_alpha1"], []),
            # =* glob matches only on boundaries between version parts,
            # so 1* does not match 10 (bug 560466).
            (
                "=cat/pkg-1.1*",
                ["cat/pkg-1.1-r1", "cat/pkg-1.10-r1"],
                ["cat/pkg-1.1-r1"],
            ),
            ("=cat/pkg-1-r1*", ["cat/pkg-1-r11"], []),
            ("=cat/pkg-1_pre*", ["cat/pkg-1_pre1"], ["cat/pkg-1_pre1"]),
            ("=cat/pkg-1-r1*", ["cat/pkg-1-r1"], ["cat/pkg-1-r1"]),
            ("=cat/pkg-1-r11*", ["cat/pkg-1-r11"], ["cat/pkg-1-r11"]),
            ("=cat/pkg-1-r11*", ["cat/pkg-01-r11"], ["cat/pkg-01-r11"]),
            ("=cat/pkg-01-r11*", ["cat/pkg-1-r11"], ["cat/pkg-1-r11"]),
            ("=cat/pkg-01-r11*", ["cat/pkg-001-r11"], ["cat/pkg-001-r11"]),
            ("=sys-fs/udev-1*", ["sys-fs/udev-123", "sys-fs/udev-123-r1"], []),
            ("=sys-fs/udev-123*", ["sys-fs/udev-123"], ["sys-fs/udev-123"]),
            ("=sys-fs/udev-123*", ["sys-fs/udev-123-r1"], ["sys-fs/udev-123-r1"]),
            ("=sys-fs/udev-4*", ["sys-fs/udev-456", "sys-fs/udev-456-r1"], []),
            ("=sys-fs/udev-456*", ["sys-fs/udev-456"], ["sys-fs/udev-456"]),
            ("*/*", ["sys-fs/udev-456"], ["sys-fs/udev-456"]),
            ("*/*:0", ["sys-fs/udev-456:0"], ["sys-fs/udev-456:0"]),
            ("*/*:1", ["sys-fs/udev-456:0"], []),
            ("sys-fs/*", ["sys-fs/udev-456"], ["sys-fs/udev-456"]),
            ("*/udev", ["sys-fs/udev-456"], ["sys-fs/udev-456"]),
            (
                "=sys-apps/portage-2*",
                ["sys-apps/portage-2.1"],
                ["sys-apps/portage-2.1"],
            ),
            (
                "=sys-apps/portage-2.1*",
                ["sys-apps/portage-2.1.2"],
                ["sys-apps/portage-2.1.2"],
            ),
            ("dev-libs/*", ["sys-apps/portage-2.1.2"], []),
            ("*/tar", ["sys-apps/portage-2.1.2"], []),
            ("*/*", ["dev-libs/A-1", "dev-libs/B-1"], ["dev-libs/A-1", "dev-libs/B-1"]),
            ("dev-libs/*", ["dev-libs/A-1", "sci-libs/B-1"], ["dev-libs/A-1"]),
            (
                "dev-libs/A[foo]",
                [Package("=dev-libs/A-1[foo]"), Package("=dev-libs/A-2[-foo]")],
                ["dev-libs/A-1"],
            ),
            (
                "dev-libs/A[-foo]",
                [Package("=dev-libs/A-1[foo]"), Package("=dev-libs/A-2[-foo]")],
                ["dev-libs/A-2"],
            ),
            (
                "dev-libs/A[-foo]",
                [Package("=dev-libs/A-1[foo]"), Package("=dev-libs/A-2")],
                [],
            ),
            (
                "dev-libs/A[foo,bar]",
                [Package("=dev-libs/A-1[foo]"), Package("=dev-libs/A-2[-foo]")],
                [],
            ),
            (
                "dev-libs/A[foo,bar]",
                [Package("=dev-libs/A-1[foo]"), Package("=dev-libs/A-2[-foo,bar]")],
                [],
            ),
            (
                "dev-libs/A[foo,bar]",
                [Package("=dev-libs/A-1[foo]"), Package("=dev-libs/A-2[foo,bar]")],
                ["dev-libs/A-2"],
            ),
            (
                "dev-libs/A[foo,bar(+)]",
                [Package("=dev-libs/A-1[-foo]"), Package("=dev-libs/A-2[foo]")],
                ["dev-libs/A-2"],
            ),
            (
                "dev-libs/A[foo,bar(-)]",
                [Package("=dev-libs/A-1[-foo]"), Package("=dev-libs/A-2[foo]")],
                [],
            ),
            (
                "dev-libs/A[foo,-bar(-)]",
                [Package("=dev-libs/A-1[-foo,bar]"), Package("=dev-libs/A-2[foo]")],
                ["dev-libs/A-2"],
            ),
            (
                "dev-libs/A::repo1",
                [Package("=dev-libs/A-1::repo1"), Package("=dev-libs/A-1::repo2")],
                ["dev-libs/A-1::repo1"],
            ),
            (
                "dev-libs/A::repo2",
                [Package("=dev-libs/A-1::repo1"), Package("=dev-libs/A-1::repo2")],
                ["dev-libs/A-1::repo2"],
            ),
            (
                "dev-libs/A::repo2[foo]",
                [
                    Package("=dev-libs/A-1::repo1[foo]"),
                    Package("=dev-libs/A-1::repo2[-foo]"),
                ],
                [],
            ),
            (
                "dev-libs/A::repo2[foo]",
                [
                    Package("=dev-libs/A-1::repo1[-foo]"),
                    Package("=dev-libs/A-1::repo2[foo]"),
                ],
                ["dev-libs/A-1::repo2"],
            ),
            (
                "dev-libs/A:1::repo2[foo]",
                [Package("=dev-libs/A-1:1::repo1"), Package("=dev-libs/A-1:2::repo2")],
                [],
            ),
            (
                "dev-libs/A:1::repo2[foo]",
                [
                    Package("=dev-libs/A-1:2::repo1"),
                    Package("=dev-libs/A-1:1::repo2[foo]"),
                ],
                ["dev-libs/A-1::repo2"],
            ),
            (
                "virtual/ffmpeg:0/53",
                [Package("=virtual/ffmpeg-0.10.3:0/53")],
                ["virtual/ffmpeg-0.10.3"],
            ),
            (
                "virtual/ffmpeg:0/53=",
                [Package("=virtual/ffmpeg-0.10.3:0/53")],
                ["virtual/ffmpeg-0.10.3"],
            ),
            ("virtual/ffmpeg:0/52", [Package("=virtual/ffmpeg-0.10.3:0/53")], []),
            (
                "virtual/ffmpeg:=",
                [Package("=virtual/ffmpeg-0.10.3:0/53")],
                ["virtual/ffmpeg-0.10.3"],
            ),
            (
                "virtual/ffmpeg:0=",
                [Package("=virtual/ffmpeg-0.10.3:0/53")],
                ["virtual/ffmpeg-0.10.3"],
            ),
            (
                "virtual/ffmpeg:*",
                [Package("=virtual/ffmpeg-0.10.3:0/53")],
                ["virtual/ffmpeg-0.10.3"],
            ),
            (
                "virtual/ffmpeg:0",
                [Package("=virtual/ffmpeg-0.10.3:0/53")],
                ["virtual/ffmpeg-0.10.3"],
            ),
            (
                "sys-libs/db:4.8/4.8",
                [Package("=sys-libs/db-4.8.30:4.8")],
                ["sys-libs/db-4.8.30"],
            ),
            (
                "sys-libs/db:4.8/4.8=",
                [Package("=sys-libs/db-4.8.30:4.8")],
                ["sys-libs/db-4.8.30"],
            ),
            (
                "sys-libs/db:4.8=",
                [Package("=sys-libs/db-4.8.30:4.8")],
                ["sys-libs/db-4.8.30"],
            ),
            (
                "sys-libs/db:*",
                [Package("=sys-libs/db-4.8.30:4.8")],
                ["sys-libs/db-4.8.30"],
            ),
            ("sys-libs/db:4.8/0", [Package("=sys-libs/db-4.8.30:4.8")], []),
            ("sys-libs/db:4.8/0=", [Package("=sys-libs/db-4.8.30:4.8")], []),
        )

        for atom, cpv_list, expected_result in tests:
            result = []
            for pkg in match_from_list(atom, cpv_list):
                if isinstance(pkg, Package):
                    if pkg.repo:
                        result.append(pkg.cpv + _repo_separator + pkg.repo)
                    else:
                        result.append(pkg.cpv)
                else:
                    result.append(pkg)
            self.assertEqual(result, expected_result)



# Copyright 2011-2014 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

import subprocess

import portage
from portage import os
from portage.const import PORTAGE_PYM_PATH
from portage.tests import TestCase


class WhirlpoolTestCase(TestCase):
    def testBundledWhirlpool(self):
        # execute the tests bundled with the whirlpool module
        retval = subprocess.call(
            [
                portage._python_interpreter,
                "-b",
                "-Wd",
                os.path.join(PORTAGE_PYM_PATH, "portage/util/whirlpool.py"),
            ]
        )
        self.assertEqual(retval, os.EX_OK)



# Copyright 2014 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

from portage.tests import TestCase
from portage.tests.resolver.ResolverPlayground import (
    ResolverPlayground,
    ResolverPlaygroundTestCase,
)


class OnlydepsTestCase(TestCase):
    def testOnlydeps(self):
        ebuilds = {
            "dev-libs/A-1": {"DEPEND": "dev-libs/B"},
            "dev-libs/B-1": {},
        }
        installed = {
            "dev-libs/B-1": {},
        }

        test_cases = (
            ResolverPlaygroundTestCase(
                ["dev-libs/A", "dev-libs/B"],
                all_permutations=True,
                success=True,
                options={"--onlydeps": True},
                mergelist=["dev-libs/B-1"],
            ),
        )

        playground = ResolverPlayground(
            ebuilds=ebuilds, installed=installed, debug=False
        )
        try:
            for test_case in test_cases:
                playground.run_TestCase(test_case)
                self.assertEqual(test_case.test_success, True, test_case.fail_msg)
        finally:
            playground.cleanup()



# test_NewsItem.py -- Portage Unit Testing Functionality
# Copyright 2007-2019 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

from portage import os
from portage.tests import TestCase
from portage.news import NewsItem
from portage.dbapi.virtual import testdbapi
from tempfile import mkstemp

# TODO(antarus) Make newsitem use a loader so we can load using a string instead of a tempfile


class NewsItemTestCase(TestCase):
    """These tests suck: they use your running config instead of making their own"""

    fakeItem = """
Title: YourSQL Upgrades from 4.0 to 4.1
Author: Ciaran McCreesh <ciaranm@gentoo.org>
Content-Type: text/plain
Posted: 01-Nov-2005
Revision: 1
News-Item-Format: 1.0
#Display-If-Installed:
#Display-If-Profile:
#Display-If-Arch:

YourSQL databases created using YourSQL version 4.0 are incompatible
with YourSQL version 4.1 or later. There is no reliable way to
automate the database format conversion, so action from the system
administrator is required before an upgrade can take place.

Please see the Gentoo YourSQL Upgrade Guide for instructions:

    http://www.gentoo.org/doc/en/yoursql-upgrading.xml

Also see the official YourSQL documentation:

    http://dev.yoursql.com/doc/refman/4.1/en/upgrading-from-4-0.html

After upgrading, you should also recompile any packages which link
against YourSQL:

    revdep-rebuild --library=libyoursqlclient.so.12

The revdep-rebuild tool is provided by app-portage/gentoolkit.
"""

    def setUp(self):
        self.profile = "/var/db/repos/gentoo/profiles/default-linux/x86/2007.0/"
        self.keywords = "x86"
        # Use fake/test dbapi to avoid slow tests
        self.vardb = testdbapi()
        # self.vardb.inject_cpv('sys-apps/portage-2.0', { 'SLOT' : 0 })
        # Consumers only use ARCH, so avoid portage.settings by using a dict
        self.settings = {"ARCH": "x86"}

    def testDisplayIfProfile(self):
        tmpItem = self.fakeItem[:].replace(
            "#Display-If-Profile:", "Display-If-Profile: %s" % self.profile
        )

        item = self._processItem(tmpItem)
        try:
            self.assertTrue(
                item.isRelevant(self.vardb, self.settings, self.profile),
                msg="Expected %s to be relevant, but it was not!" % tmpItem,
            )
        finally:
            os.unlink(item.path)

    def testDisplayIfInstalled(self):
        tmpItem = self.fakeItem[:].replace(
            "#Display-If-Installed:", "Display-If-Installed: %s" % "sys-apps/portage"
        )

        try:
            item = self._processItem(tmpItem)
            self.assertTrue(
                item.isRelevant(self.vardb, self.settings, self.profile),
                msg="Expected %s to be relevant, but it was not!" % tmpItem,
            )
        finally:
            os.unlink(item.path)

    def testDisplayIfKeyword(self):
        tmpItem = self.fakeItem[:].replace(
            "#Display-If-Keyword:", "Display-If-Keyword: %s" % self.keywords
        )

        try:
            item = self._processItem(tmpItem)
            self.assertTrue(
                item.isRelevant(self.vardb, self.settings, self.profile),
                msg="Expected %s to be relevant, but it was not!" % tmpItem,
            )
        finally:
            os.unlink(item.path)

    def _processItem(self, item):
        filename = None
        fd, filename = mkstemp()
        f = os.fdopen(fd, "w")
        f.write(item)
        f.close()
        try:
            return NewsItem(filename, 0)
        except TypeError:
            self.fail("Error while processing news item %s" % filename)



# Copyright 2012 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

try:
    import ctypes
    import ctypes.util
except ImportError:
    ctypes = None
else:
    try:
        ctypes.cdll
    except AttributeError:
        ctypes = None

_library_names = {}


def find_library(name):
    """
    Calls ctype.util.find_library() if the ctypes module is available,
    and otherwise returns None. Results are cached for future invocations.
    """
    filename = _library_names.get(name)
    if filename is None:
        if ctypes is not None:
            filename = ctypes.util.find_library(name)
            if filename is None:
                filename = False
            _library_names[name] = filename

    if filename is False:
        return None
    return filename


_library_handles = {}


def LoadLibrary(name):
    """
    Calls ctypes.cdll.LoadLibrary(name) if the ctypes module is available,
    and otherwise returns None. Results are cached for future invocations.
    """
    handle = _library_handles.get(name)

    if handle is None and ctypes is not None:
        handle = ctypes.CDLL(name, use_errno=True)
        _library_handles[name] = handle

    return handle



# Copyright 2017 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

from portage.tests import TestCase
from portage.tests.resolver.ResolverPlayground import (
    ResolverPlayground,
    ResolverPlaygroundTestCase,
)


class SlotOperatorExclusiveSlotsTestCase(TestCase):
    def testSlotOperatorExclusiveSlots(self):

        ebuilds = {
            "media-libs/mesa-17.0.1": {
                "EAPI": "6",
                "SLOT": "0",
                "RDEPEND": "<sys-devel/llvm-5:=",
            },
            "sys-devel/clang-4.0.0": {
                "EAPI": "6",
                "SLOT": "4",
                "RDEPEND": (
                    "~sys-devel/llvm-4.0.0:4= " "!sys-devel/llvm:0 !sys-devel/clang:0"
                ),
            },
            "sys-devel/clang-3.9.1-r100": {
                "EAPI": "6",
                "SLOT": "0/3.9.1",
                "RDEPEND": "~sys-devel/llvm-3.9.1",
            },
            "sys-devel/llvm-4.0.0": {
                "EAPI": "6",
                "SLOT": "4",
                "RDEPEND": "!sys-devel/llvm:0",
            },
            "sys-devel/llvm-3.9.1": {
                "EAPI": "6",
                "SLOT": "0/3.91",
                "RDEPEND": "!sys-devel/llvm:0",
                "PDEPEND": "=sys-devel/clang-3.9.1-r100",
            },
        }

        installed = {
            "media-libs/mesa-17.0.1": {
                "EAPI": "6",
                "SLOT": "0",
                "RDEPEND": "<sys-devel/llvm-5:0/3.9.1=",
            },
            "sys-devel/clang-3.9.1-r100": {
                "EAPI": "6",
                "SLOT": "0/3.9.1",
                "RDEPEND": "~sys-devel/llvm-3.9.1",
            },
            "sys-devel/llvm-3.9.1": {
                "EAPI": "6",
                "SLOT": "0/3.9.1",
                "RDEPEND": "!sys-devel/llvm:0",
                "PDEPEND": "=sys-devel/clang-3.9.1-r100",
            },
        }

        world = ["sys-devel/clang", "media-libs/mesa"]

        test_cases = (
            # Test bug #612772, where slot operator rebuilds are not
            # properly triggered (for things like mesa) during a
            # llvm:0 to llvm:4 upgrade with clang, resulting in
            # unsolved blockers.
            ResolverPlaygroundTestCase(
                ["@world"],
                options={"--update": True, "--deep": True},
                success=True,
                ambiguous_merge_order=True,
                mergelist=[
                    "sys-devel/llvm-4.0.0",
                    "media-libs/mesa-17.0.1",
                    (
                        "sys-devel/clang-4.0.0",
                        "[uninstall]sys-devel/llvm-3.9.1",
                        "!sys-devel/llvm:0",
                        "[uninstall]sys-devel/clang-3.9.1-r100",
                        "!sys-devel/clang:0",
                    ),
                ],
            ),
        )

        playground = ResolverPlayground(
            ebuilds=ebuilds, installed=installed, world=world
        )
        try:
            for test_case in test_cases:
                playground.run_TestCase(test_case)
                self.assertEqual(test_case.test_success, True, test_case.fail_msg)
        finally:
            playground.cleanup()

        world = ["media-libs/mesa"]

        test_cases = (
            # Test bug #612874, where a direct circular dependency
            # between llvm-3.9.1 and clang-3.9.1-r100 causes a
            # missed update from llvm:0 to llvm:4. Since llvm:4 does
            # not have a dependency on clang, the upgrade from llvm:0
            # to llvm:4 makes the installed sys-devel/clang-3.9.1-r100
            # instance eligible for removal by emerge --depclean, which
            # explains why clang does not appear in the mergelist.
            ResolverPlaygroundTestCase(
                ["@world"],
                options={"--update": True, "--deep": True},
                success=True,
                ambiguous_merge_order=True,
                mergelist=[
                    "sys-devel/llvm-4.0.0",
                    (
                        "media-libs/mesa-17.0.1",
                        "[uninstall]sys-devel/llvm-3.9.1",
                        "!sys-devel/llvm:0",
                    ),
                ],
            ),
        )

        playground = ResolverPlayground(
            ebuilds=ebuilds, installed=installed, world=world
        )
        try:
            for test_case in test_cases:
                playground.run_TestCase(test_case)
                self.assertEqual(test_case.test_success, True, test_case.fail_msg)
        finally:
            playground.cleanup()



# Copyright 2010-2020 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

from portage.tests import TestCase
from portage.tests.resolver.ResolverPlayground import (
    ResolverPlayground,
    ResolverPlaygroundTestCase,
)


class MissingIUSEandEvaluatedAtomsTestCase(TestCase):
    def testMissingIUSEandEvaluatedAtoms(self):
        ebuilds = {
            "dev-libs/A-1": {
                "DEPEND": "dev-libs/B[foo?]",
                "IUSE": "foo bar",
                "EAPI": 2,
            },
            "dev-libs/A-2": {
                "DEPEND": "dev-libs/B[foo?,bar]",
                "IUSE": "foo bar",
                "EAPI": 2,
            },
            "dev-libs/B-1": {"IUSE": "bar"},
        }

        test_cases = (
            ResolverPlaygroundTestCase(["=dev-libs/A-1"], success=False),
            ResolverPlaygroundTestCase(["=dev-libs/A-2"], success=False),
        )

        playground = ResolverPlayground(ebuilds=ebuilds, debug=False)
        try:
            for test_case in test_cases:
                playground.run_TestCase(test_case)
                self.assertEqual(test_case.test_success, True, test_case.fail_msg)
        finally:
            playground.cleanup()



# Copyright 2015 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

import sys

from portage.const import SUPPORTED_GENTOO_BINPKG_FORMATS
from portage.tests import TestCase
from portage.tests.resolver.ResolverPlayground import (
    ResolverPlayground,
    ResolverPlaygroundTestCase,
)
from portage.output import colorize


class SonameReinstallTestCase(TestCase):
    def testSonameReinstall(self):

        binpkgs = {
            "app-misc/A-1": {
                "RDEPEND": "dev-libs/B",
                "DEPEND": "dev-libs/B",
                "REQUIRES": "x86_32: libB.so.2",
            },
            "dev-libs/B-2": {
                "PROVIDES": "x86_32: libB.so.2",
            },
            "dev-libs/B-1": {
                "PROVIDES": "x86_32: libB.so.1",
            },
        }

        installed = {
            "app-misc/A-1": {
                "RDEPEND": "dev-libs/B",
                "DEPEND": "dev-libs/B",
                "REQUIRES": "x86_32: libB.so.1",
            },
            "dev-libs/B-1": {
                "PROVIDES": "x86_32: libB.so.1",
            },
        }

        world = ("app-misc/A",)

        test_cases = (
            # Test that --ignore-soname-deps prevents the above
            # rebuild from being triggered.
            ResolverPlaygroundTestCase(
                ["@world"],
                options={
                    "--deep": True,
                    "--ignore-soname-deps": "n",
                    "--update": True,
                    "--usepkgonly": True,
                },
                success=True,
                mergelist=[
                    "[binary]dev-libs/B-2",
                    "[binary]app-misc/A-1",
                ],
            ),
            # Test that --ignore-soname-deps prevents the above
            # reinstall from being triggered.
            ResolverPlaygroundTestCase(
                ["@world"],
                options={
                    "--deep": True,
                    "--ignore-soname-deps": "y",
                    "--update": True,
                    "--usepkgonly": True,
                },
                success=True,
                mergelist=[
                    "[binary]dev-libs/B-2",
                ],
            ),
        )

        for binpkg_format in SUPPORTED_GENTOO_BINPKG_FORMATS:
            with self.subTest(binpkg_format=binpkg_format):
                print(colorize("HILITE", binpkg_format), end=" ... ")
                sys.stdout.flush()
                playground = ResolverPlayground(
                    debug=False,
                    binpkgs=binpkgs,
                    installed=installed,
                    world=world,
                    user_config={
                        "make.conf": ('BINPKG_FORMAT="%s"' % binpkg_format,),
                    },
                )

                try:
                    for test_case in test_cases:
                        playground.run_TestCase(test_case)
                        self.assertEqual(
                            test_case.test_success, True, test_case.fail_msg
                        )
                finally:
                    # Disable debug so that cleanup works.
                    playground.debug = False
                    playground.cleanup()



# Copyright 2005-2012 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

doc = """Perform package move updates for installed and binary packages."""
__doc__ = doc


module_spec = {
    "name": "move",
    "description": doc,
    "provides": {
        "module1": {
            "name": "moveinst",
            "sourcefile": "move",
            "class": "MoveInstalled",
            "description": doc,
            "options": ["check", "fix"],
            "functions": ["check", "fix"],
            "func_desc": {},
        },
        "module2": {
            "name": "movebin",
            "sourcefile": "move",
            "class": "MoveBinary",
            "description": "Perform package move updates for binary packages",
            "functions": ["check", "fix"],
            "func_desc": {},
        },
    },
}



# Copyright 2020 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

from portage.tests import TestCase
from portage.tests.resolver.ResolverPlayground import (
    ResolverPlayground,
    ResolverPlaygroundTestCase,
)


class CircularRustTestCase(TestCase):
    def testCircularPypyExe(self):

        ebuilds = {
            "dev-lang/rust-1.47.0-r2": {
                "EAPI": "7",
                "SLOT": "stable/1.47",
                "BDEPEND": "|| ( =dev-lang/rust-1.46* =dev-lang/rust-bin-1.46* =dev-lang/rust-1.47* =dev-lang/rust-bin-1.47* )",
            },
            "dev-lang/rust-1.46.0": {
                "EAPI": "7",
                "SLOT": "stable/1.46",
                "BDEPEND": "|| ( =dev-lang/rust-1.45* =dev-lang/rust-bin-1.45* =dev-lang/rust-1.46* =dev-lang/rust-bin-1.46* )",
            },
            "dev-lang/rust-bin-1.47.0": {
                "EAPI": "7",
            },
            "dev-lang/rust-bin-1.46.0": {
                "EAPI": "7",
            },
        }

        installed = {
            "dev-lang/rust-1.46.0": {
                "EAPI": "7",
                "SLOT": "stable/1.46",
                "BDEPEND": "|| ( =dev-lang/rust-1.45* =dev-lang/rust-bin-1.45* =dev-lang/rust-1.46* =dev-lang/rust-bin-1.46* )",
            },
        }

        test_cases = (
            # Test bug 756961, where a circular dependency was reported
            # when a package would replace its own builtime dependency.
            # This needs to be tested with and without --update, since
            # that affects package selection logic significantly,
            # expecially for packages given as arguments.
            ResolverPlaygroundTestCase(
                ["=dev-lang/rust-1.46*"],
                mergelist=["dev-lang/rust-1.46.0"],
                success=True,
            ),
            ResolverPlaygroundTestCase(
                ["=dev-lang/rust-1.46*"],
                options={"--update": True},
                mergelist=[],
                success=True,
            ),
            ResolverPlaygroundTestCase(
                ["=dev-lang/rust-1.46*"],
                options={"--deep": True, "--update": True},
                mergelist=[],
                success=True,
            ),
            ResolverPlaygroundTestCase(
                ["dev-lang/rust"],
                mergelist=["dev-lang/rust-1.47.0-r2"],
                success=True,
            ),
            ResolverPlaygroundTestCase(
                ["dev-lang/rust"],
                options={"--update": True},
                mergelist=["dev-lang/rust-1.47.0-r2"],
                success=True,
            ),
            ResolverPlaygroundTestCase(
                ["@world"],
                options={"--deep": True, "--update": True},
                mergelist=["dev-lang/rust-1.47.0-r2"],
                success=True,
            ),
        )

        world = ["dev-lang/rust"]

        playground = ResolverPlayground(
            ebuilds=ebuilds, installed=installed, world=world, debug=False
        )
        try:
            for test_case in test_cases:
                playground.run_TestCase(test_case)
                self.assertEqual(test_case.test_success, True, test_case.fail_msg)
        finally:
            playground.debug = False
            playground.cleanup()



# Copyright 2010-2020 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

"""Provides an easy-to-use python interface to Gentoo's metadata.xml file.

	Example usage:
		>>> from portage.xml.metadata import MetaDataXML
		>>> pkg_md = MetaDataXML('/var/db/repos/gentoo/app-misc/gourmet/metadata.xml')
		>>> pkg_md
		<MetaDataXML '/var/db/repos/gentoo/app-misc/gourmet/metadata.xml'>
		>>> pkg_md.herds()
		['no-herd']
		>>> for maint in pkg_md.maintainers():
		...     print "{0} ({1})".format(maint.email, maint.name)
		...
		nixphoeni@gentoo.org (Joe Sapp)
		>>> for flag in pkg_md.use():
		...     print flag.name, "->", flag.description
		...
		rtf -> Enable export to RTF
		gnome-print -> Enable printing support using gnome-print
		>>> upstream = pkg_md.upstream()
		>>> upstream
		[<_Upstream {'docs': [], 'remoteid': [], 'maintainer':
		 [<_Maintainer 'Thomas_Hinkle@alumni.brown.edu'>], 'bugtracker': [],
		 'changelog': []}>]
		>>> upstream[0].maintainer[0].name
		'Thomas Mills Hinkle'
"""

__all__ = ("MetaDataXML", "parse_metadata_use")


import re
import xml.etree.ElementTree as etree

try:
    from xml.parsers.expat import ExpatError
except Exception:
    ExpatError = SyntaxError

from portage import _encodings, _unicode_encode
from portage.util import cmp_sort_key, unique_everseen


class _MetadataTreeBuilder(etree.TreeBuilder):
    """
    Implements doctype() as required to avoid deprecation warnings with
    Python >=2.7.
    """

    def doctype(self, name, pubid, system):
        pass


class _Maintainer:
    """An object for representing one maintainer.

    @type email: str or None
    @ivar email: Maintainer's email address. Used for both Gentoo and upstream.
    @type name: str or None
    @ivar name: Maintainer's name. Used for both Gentoo and upstream.
    @type description: str or None
    @ivar description: Description of what a maintainer does. Gentoo only.
    @type maint_type: str or None
    @ivar maint_type: GLEP67 maintainer type (project or person). Gentoo only.
    @type restrict: str or None
    @ivar restrict: e.g. &gt;=portage-2.2 means only maintains versions
            of Portage greater than 2.2. Should be DEPEND string with < and >
            converted to &lt; and &gt; respectively.
    @type status: str or None
    @ivar status: If set, either 'active' or 'inactive'. Upstream only.
    """

    def __init__(self, node):
        self.email = None
        self.name = None
        self.description = None
        self.maint_type = node.get("type")
        self.restrict = node.get("restrict")
        self.status = node.get("status")
        for attr in node:
            setattr(self, attr.tag, attr.text)

    def __repr__(self):
        return "<{} {!r}>".format(self.__class__.__name__, self.email)


class _Useflag:
    """An object for representing one USE flag.

    @todo: Is there any way to have a keyword option to leave in
            <pkg> and <cat> for later processing?
    @type name: str or None
    @ivar name: USE flag
    @type restrict: str or None
    @ivar restrict: e.g. &gt;=portage-2.2 means flag is only available in
            versions greater than 2.2
    @type description: str
    @ivar description: description of the USE flag
    """

    def __init__(self, node):
        self.name = node.get("name")
        self.restrict = node.get("restrict")
        _desc = ""
        if node.text:
            _desc = node.text
        for child in node:
            _desc += child.text if child.text else ""
            _desc += child.tail if child.tail else ""
        # This takes care of tabs and newlines left from the file
        self.description = re.sub(r"\s+", " ", _desc)

    def __repr__(self):
        return "<{} {!r}>".format(self.__class__.__name__, self.name)


class _Upstream:
    """An object for representing one package's upstream.

    @type maintainers: list
    @ivar maintainers: L{_Maintainer} objects for each upstream maintainer
    @type changelogs: list
    @ivar changelogs: URLs to upstream's ChangeLog file in str format
    @type docs: list
    @ivar docs: Sequence of tuples containing URLs to upstream documentation
            in the first slot and 'lang' attribute in the second, e.g.,
            [('http.../docs/en/tut.html', None), ('http.../doc/fr/tut.html', 'fr')]
    @type bugtrackers: list
    @ivar bugtrackers: URLs to upstream's bugtracker. May also contain an email
            address if prepended with 'mailto:'
    @type remoteids: list
    @ivar remoteids: Sequence of tuples containing the project's hosting site
            name in the first slot and the project's ID name or number for that
            site in the second, e.g., [('sourceforge', 'systemrescuecd')]
    """

    def __init__(self, node):
        self.node = node
        self.maintainers = self.upstream_maintainers()
        self.changelogs = self.upstream_changelogs()
        self.docs = self.upstream_documentation()
        self.bugtrackers = self.upstream_bugtrackers()
        self.remoteids = self.upstream_remoteids()

    def __repr__(self):
        return "<{} {!r}>".format(self.__class__.__name__, self.__dict__)

    def upstream_bugtrackers(self):
        """Retrieve upstream bugtracker location from xml node."""
        return [e.text for e in self.node.findall("bugs-to") if e.text]

    def upstream_changelogs(self):
        """Retrieve upstream changelog location from xml node."""
        return [e.text for e in self.node.findall("changelog") if e.text]

    def upstream_documentation(self):
        """Retrieve upstream documentation location from xml node."""
        result = []
        for elem in (e for e in self.node.findall("doc") if e.text):
            lang = elem.get("lang")
            result.append((elem.text, lang))
        return result

    def upstream_maintainers(self):
        """Retrieve upstream maintainer information from xml node."""
        return [_Maintainer(m) for m in self.node.findall("maintainer")]

    def upstream_remoteids(self):
        """Retrieve upstream remote ID from xml node."""
        return [
            (e.text, e.get("type")) for e in self.node.findall("remote-id") if e.text
        ]


class MetaDataXML:
    """Access metadata.xml"""

    def __init__(self, metadata_xml_path, herds):
        """Parse a valid metadata.xml file.

        @type metadata_xml_path: str
        @param metadata_xml_path: path to a valid metadata.xml file
        @type herds: str or ElementTree
        @param herds: path to a herds.xml, or a pre-parsed ElementTree
        @raise IOError: if C{metadata_xml_path} can not be read
        """

        self.metadata_xml_path = metadata_xml_path
        self._xml_tree = None

        try:
            self._xml_tree = etree.parse(
                _unicode_encode(
                    metadata_xml_path, encoding=_encodings["fs"], errors="strict"
                ),
                parser=etree.XMLParser(target=_MetadataTreeBuilder()),
            )
        except ImportError:
            pass
        except ExpatError as e:
            raise SyntaxError("{}".format(e))

        if isinstance(herds, etree.ElementTree):
            herds_etree = herds
            herds_path = None
        else:
            herds_etree = None
            herds_path = herds

        # Used for caching
        self._herdstree = herds_etree
        self._herds_path = herds_path
        self._descriptions = None
        self._maintainers = None
        self._herds = None
        self._useflags = None
        self._upstream = None

    def __repr__(self):
        return "<{} {!r}>".format(self.__class__.__name__, self.metadata_xml_path)

    def _get_herd_email(self, herd):
        """Get a herd's email address.

        @type herd: str
        @param herd: herd whose email you want
        @rtype: str or None
        @return: email address or None if herd is not in herds.xml
        @raise IOError: if $PORTDIR/metadata/herds.xml can not be read
        """

        if self._herdstree is None:
            try:
                self._herdstree = etree.parse(
                    _unicode_encode(
                        self._herds_path, encoding=_encodings["fs"], errors="strict"
                    ),
                    parser=etree.XMLParser(target=_MetadataTreeBuilder()),
                )
            except (ImportError, OSError, SyntaxError):
                return None

        # Some special herds are not listed in herds.xml
        if herd in ("no-herd", "maintainer-wanted", "maintainer-needed"):
            return None

        try:
            # Python 2.7 or >=3.2
            iterate = self._herdstree.iter
        except AttributeError:
            iterate = self._herdstree.getiterator

        for node in iterate("herd"):
            if node.findtext("name") == herd:
                return node.findtext("email")

    def herds(self, include_email=False):
        """Return a list of text nodes for <herd>.

        @type include_email: bool
        @keyword include_email: if True, also look up the herd's email
        @rtype: tuple
        @return: if include_email is False, return a list of strings;
                 if include_email is True, return a list of tuples containing:
                                 [('herd1', 'herd1@gentoo.org'), ('no-herd', None);
        """
        if self._herds is None:
            if self._xml_tree is None:
                self._herds = tuple()
            else:
                herds = []
                for elem in self._xml_tree.findall("herd"):
                    text = elem.text
                    if text is None:
                        text = ""
                    if include_email:
                        herd_mail = self._get_herd_email(text)
                        herds.append((text, herd_mail))
                    else:
                        herds.append(text)
                self._herds = tuple(herds)

        return self._herds

    def descriptions(self):
        """Return a list of text nodes for <longdescription>.

        @rtype: list
        @return: package description in string format
        @todo: Support the C{lang} attribute
        """
        if self._descriptions is None:
            if self._xml_tree is None:
                self._descriptions = tuple()
            else:
                self._descriptions = tuple(
                    e.text for e in self._xml_tree.findall("longdescription") if e.text
                )

        return self._descriptions

    def maintainers(self):
        """Get maintainers' name, email and description.

        @rtype: list
        @return: a sequence of L{_Maintainer} objects in document order.
        """

        if self._maintainers is None:
            if self._xml_tree is None:
                self._maintainers = tuple()
            else:
                self._maintainers = tuple(
                    _Maintainer(node) for node in self._xml_tree.findall("maintainer")
                )

        return self._maintainers

    def use(self):
        """Get names and descriptions for USE flags defined in metadata.

        @rtype: list
        @return: a sequence of L{_Useflag} objects in document order.
        """

        if self._useflags is None:
            if self._xml_tree is None:
                self._useflags = tuple()
            else:
                try:
                    # Python 2.7 or >=3.2
                    iterate = self._xml_tree.iter
                except AttributeError:
                    iterate = self._xml_tree.getiterator
                self._useflags = tuple(_Useflag(node) for node in iterate("flag"))

        return self._useflags

    def upstream(self):
        """Get upstream contact information.

        @rtype: list
        @return: a sequence of L{_Upstream} objects in document order.
        """

        if self._upstream is None:
            if self._xml_tree is None:
                self._upstream = tuple()
            else:
                self._upstream = tuple(
                    _Upstream(node) for node in self._xml_tree.findall("upstream")
                )

        return self._upstream

    def format_maintainer_string(self):
        """Format string containing maintainers and herds (emails if possible).
        Used by emerge to display maintainer information.
        Entries are sorted according to the rules stated on the bug wranglers page.

        @rtype: String
        @return: a string containing maintainers and herds
        """
        maintainers = []
        for maintainer in self.maintainers():
            if maintainer.email is None or not maintainer.email.strip():
                if maintainer.name and maintainer.name.strip():
                    maintainers.append(maintainer.name)
            else:
                maintainers.append(maintainer.email)

        for herd, email in self.herds(include_email=True):
            if herd == "no-herd":
                continue
            if email is None or not email.strip():
                if herd and herd.strip():
                    maintainers.append(herd)
            else:
                maintainers.append(email)

        maintainers = list(unique_everseen(maintainers))

        maint_str = ""
        if maintainers:
            maint_str = maintainers[0]
            maintainers = maintainers[1:]
        if maintainers:
            maint_str += " " + ",".join(maintainers)

        return maint_str

    def format_upstream_string(self):
        """Format string containing upstream maintainers and bugtrackers.
        Used by emerge to display upstream information.

        @rtype: String
        @return: a string containing upstream maintainers and bugtrackers
        """
        maintainers = []
        for upstream in self.upstream():
            for maintainer in upstream.maintainers:
                if maintainer.email is None or not maintainer.email.strip():
                    if maintainer.name and maintainer.name.strip():
                        maintainers.append(maintainer.name)
                else:
                    maintainers.append(maintainer.email)

            for bugtracker in upstream.bugtrackers:
                if bugtracker.startswith("mailto:"):
                    bugtracker = bugtracker[7:]
                maintainers.append(bugtracker)

        maintainers = list(unique_everseen(maintainers))
        maint_str = " ".join(maintainers)
        return maint_str


# lang with higher value is preferred
_lang_pref = {
    "": 0,
    "en": 1,
}


def _cmp_lang(a, b):
    a_score = _lang_pref.get(a.get("lang", ""), -1)
    b_score = _lang_pref.get(b.get("lang", ""), -1)

    return a_score - b_score


def parse_metadata_use(xml_tree):
    """
    Records are wrapped in XML as per GLEP 56
    returns a dict with keys constisting of USE flag names and values
    containing their respective descriptions
    """
    uselist = {}

    usetags = xml_tree.findall("use")
    if not usetags:
        return uselist

    # Sort by language preference in descending order.
    usetags.sort(key=cmp_sort_key(_cmp_lang), reverse=True)

    # It's possible to have multiple 'use' elements.
    for usetag in usetags:
        flags = usetag.findall("flag")
        if not flags:
            # DTD allows use elements containing no flag elements.
            continue

        for flag in flags:
            pkg_flag = flag.get("name")
            if pkg_flag is not None:
                flag_restrict = flag.get("restrict")

                # Descriptions may exist for multiple languages, so
                # ignore all except the first description found for a
                # particular value of restrict (see bug 599060).
                try:
                    uselist[pkg_flag][flag_restrict]
                except KeyError:
                    pass
                else:
                    continue

                # emulate the Element.itertext() method from python-2.7
                inner_text = []
                stack = []
                stack.append(flag)
                while stack:
                    obj = stack.pop()
                    if isinstance(obj, str):
                        inner_text.append(obj)
                        continue
                    if isinstance(obj.text, str):
                        inner_text.append(obj.text)
                    if isinstance(obj.tail, str):
                        stack.append(obj.tail)
                    stack.extend(reversed(obj))

                if flag.get("name") not in uselist:
                    uselist[flag.get("name")] = {}

                # (flag_restrict can be None)
                uselist[flag.get("name")][flag_restrict] = " ".join(
                    "".join(inner_text).split()
                )
    return uselist



# Copyright 2014 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2


from portage import os, _encodings
from portage.tests import TestCase
from portage.tests.resolver.ResolverPlayground import (
    ResolverPlayground,
    ResolverPlaygroundTestCase,
)
from portage.util import ensure_dirs


class ProfilePackageSetTestCase(TestCase):
    def testProfilePackageSet(self):

        repo_configs = {
            "test_repo": {
                "layout.conf": ("profile-formats = profile-set",),
            }
        }

        profiles = (
            (
                "default/linux",
                {
                    "eapi": ("5",),
                    "packages": (
                        "*sys-libs/A",
                        "app-misc/A",
                        "app-misc/B",
                        "app-misc/C",
                    ),
                },
            ),
            (
                "default/linux/x86",
                {"eapi": ("5",), "packages": ("-app-misc/B",), "parent": ("..",)},
            ),
        )

        ebuilds = {
            "sys-libs/A-1": {
                "EAPI": "5",
            },
            "app-misc/A-1": {
                "EAPI": "5",
            },
            "app-misc/B-1": {
                "EAPI": "5",
            },
            "app-misc/C-1": {
                "EAPI": "5",
            },
        }

        installed = {
            "sys-libs/A-1": {
                "EAPI": "5",
            },
            "app-misc/A-1": {
                "EAPI": "5",
            },
            "app-misc/B-1": {
                "EAPI": "5",
            },
            "app-misc/C-1": {
                "EAPI": "5",
            },
        }

        test_cases = (
            ResolverPlaygroundTestCase(
                ["@world"],
                options={"--update": True, "--deep": True},
                mergelist=[],
                success=True,
            ),
            ResolverPlaygroundTestCase(
                [],
                options={"--depclean": True},
                success=True,
                cleanlist=["app-misc/B-1"],
            ),
        )

        playground = ResolverPlayground(
            debug=False, ebuilds=ebuilds, installed=installed, repo_configs=repo_configs
        )
        try:
            repo_dir = playground.settings.repositories.get_location_for_name(
                "test_repo"
            )
            profile_root = os.path.join(repo_dir, "profiles")

            for p, data in profiles:
                prof_path = os.path.join(profile_root, p)
                ensure_dirs(prof_path)
                for k, v in data.items():
                    with open(
                        os.path.join(prof_path, k),
                        mode="w",
                        encoding=_encodings["repo.content"],
                    ) as f:
                        for line in v:
                            f.write("%s\n" % line)

            # The config must be reloaded in order to account
            # for the above profile customizations.
            playground.reload_config()

            for test_case in test_cases:
                playground.run_TestCase(test_case)
                self.assertEqual(test_case.test_success, True, test_case.fail_msg)

        finally:
            playground.cleanup()



# Copyright 2010-2020 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import tempfile

import portage
from portage import os, shutil, _encodings
from portage.const import USER_CONFIG_PATH
from portage.dep import Atom
from portage.package.ebuild.config import config
from portage.package.ebuild._config.LicenseManager import LicenseManager
from portage.tests import TestCase
from portage.tests.resolver.ResolverPlayground import (
    ResolverPlayground,
    ResolverPlaygroundTestCase,
)
from portage.util import normalize_path


class ConfigTestCase(TestCase):
    def testClone(self):
        """
        Test the clone via constructor.
        """

        ebuilds = {
            "dev-libs/A-1": {},
        }

        playground = ResolverPlayground(ebuilds=ebuilds)
        try:
            settings = config(clone=playground.settings)
            result = playground.run(["=dev-libs/A-1"])
            pkg, existing_node = result.depgraph._select_package(
                playground.eroot, Atom("=dev-libs/A-1")
            )
            settings.setcpv(pkg)

            # clone after setcpv tests deepcopy of LazyItemsDict
            settings2 = config(clone=settings)
        finally:
            playground.cleanup()

    def testFeaturesMutation(self):
        """
        Test whether mutation of config.features updates the FEATURES
        variable and persists through config.regenerate() calls. Also
        verify that features_set._prune_overrides() works correctly.
        """
        playground = ResolverPlayground()
        try:
            settings = config(clone=playground.settings)

            settings.features.add("noclean")
            self.assertEqual("noclean" in settings["FEATURES"].split(), True)
            settings.regenerate()
            self.assertEqual("noclean" in settings["FEATURES"].split(), True)

            settings.features.discard("noclean")
            self.assertEqual("noclean" in settings["FEATURES"].split(), False)
            settings.regenerate()
            self.assertEqual("noclean" in settings["FEATURES"].split(), False)

            settings.features.add("noclean")
            self.assertEqual("noclean" in settings["FEATURES"].split(), True)
            settings.regenerate()
            self.assertEqual("noclean" in settings["FEATURES"].split(), True)

            # before: ['noclean', '-noclean', 'noclean']
            settings.features._prune_overrides()
            #  after: ['noclean']
            self.assertEqual(settings._features_overrides.count("noclean"), 1)
            self.assertEqual(settings._features_overrides.count("-noclean"), 0)

            settings.features.remove("noclean")

            # before: ['noclean', '-noclean']
            settings.features._prune_overrides()
            #  after: ['-noclean']
            self.assertEqual(settings._features_overrides.count("noclean"), 0)
            self.assertEqual(settings._features_overrides.count("-noclean"), 1)
        finally:
            playground.cleanup()

    def testLicenseManager(self):

        user_config = {
            "package.license": (
                "dev-libs/* TEST",
                "dev-libs/A -TEST2",
                "=dev-libs/A-2 TEST3 @TEST",
                "*/* @EULA TEST2",
                "=dev-libs/C-1 *",
                "=dev-libs/C-2 -*",
            ),
        }

        playground = ResolverPlayground(user_config=user_config)
        try:
            portage.util.noiselimit = -2

            license_group_locations = (
                os.path.join(
                    playground.settings.repositories["test_repo"].location, "profiles"
                ),
            )
            pkg_license = os.path.join(playground.eroot, "etc", "portage")

            lic_man = LicenseManager(license_group_locations, pkg_license)

            self.assertEqual(lic_man._accept_license_str, None)
            self.assertEqual(lic_man._accept_license, None)
            self.assertEqual(lic_man._license_groups, {"EULA": frozenset(["TEST"])})
            self.assertEqual(lic_man._undef_lic_groups, {"TEST"})

            self.assertEqual(lic_man.extract_global_changes(), "TEST TEST2")
            self.assertEqual(lic_man.extract_global_changes(), "")

            lic_man.set_accept_license_str("TEST TEST2")
            self.assertEqual(
                lic_man._getPkgAcceptLicense("dev-libs/B-1", "0", None),
                ["TEST", "TEST2", "TEST"],
            )
            self.assertEqual(
                lic_man._getPkgAcceptLicense("dev-libs/A-1", "0", None),
                ["TEST", "TEST2", "TEST", "-TEST2"],
            )
            self.assertEqual(
                lic_man._getPkgAcceptLicense("dev-libs/A-2", "0", None),
                ["TEST", "TEST2", "TEST", "-TEST2", "TEST3", "@TEST"],
            )

            self.assertEqual(
                lic_man.get_prunned_accept_license(
                    "dev-libs/B-1", [], "TEST", "0", None
                ),
                "TEST",
            )
            self.assertEqual(
                lic_man.get_prunned_accept_license(
                    "dev-libs/A-1", [], "-TEST2", "0", None
                ),
                "",
            )
            self.assertEqual(
                lic_man.get_prunned_accept_license(
                    "dev-libs/A-2", [], "|| ( TEST TEST2 )", "0", None
                ),
                "TEST",
            )
            self.assertEqual(
                lic_man.get_prunned_accept_license(
                    "dev-libs/C-1", [], "TEST5", "0", None
                ),
                "TEST5",
            )
            self.assertEqual(
                lic_man.get_prunned_accept_license(
                    "dev-libs/C-2", [], "TEST2", "0", None
                ),
                "",
            )

            self.assertEqual(
                lic_man.getMissingLicenses("dev-libs/B-1", [], "TEST", "0", None), []
            )
            self.assertEqual(
                lic_man.getMissingLicenses("dev-libs/A-1", [], "-TEST2", "0", None),
                ["-TEST2"],
            )
            self.assertEqual(
                lic_man.getMissingLicenses(
                    "dev-libs/A-2", [], "|| ( TEST TEST2 )", "0", None
                ),
                [],
            )
            self.assertEqual(
                lic_man.getMissingLicenses(
                    "dev-libs/A-3", [], "|| ( TEST2 || ( TEST3 TEST4 ) )", "0", None
                ),
                ["TEST2", "TEST3", "TEST4"],
            )
            self.assertEqual(
                lic_man.getMissingLicenses("dev-libs/C-1", [], "TEST5", "0", None), []
            )
            self.assertEqual(
                lic_man.getMissingLicenses("dev-libs/C-2", [], "TEST2", "0", None),
                ["TEST2"],
            )
            self.assertEqual(
                lic_man.getMissingLicenses("dev-libs/D-1", [], "", "0", None), []
            )
        finally:
            portage.util.noiselimit = 0
            playground.cleanup()

    def testPackageMaskOrder(self):

        ebuilds = {
            "dev-libs/A-1": {},
            "dev-libs/B-1": {},
            "dev-libs/C-1": {},
            "dev-libs/D-1": {},
            "dev-libs/E-1": {},
        }

        repo_configs = {
            "test_repo": {
                "package.mask": (
                    "dev-libs/A",
                    "dev-libs/C",
                ),
            }
        }

        profile = {
            "package.mask": (
                "-dev-libs/A",
                "dev-libs/B",
                "-dev-libs/B",
                "dev-libs/D",
            ),
        }

        user_config = {
            "package.mask": (
                "-dev-libs/C",
                "-dev-libs/D",
                "dev-libs/E",
            ),
        }

        test_cases = (
            ResolverPlaygroundTestCase(
                ["dev-libs/A"], options={"--autounmask": "n"}, success=False
            ),
            ResolverPlaygroundTestCase(
                ["dev-libs/B"], success=True, mergelist=["dev-libs/B-1"]
            ),
            ResolverPlaygroundTestCase(
                ["dev-libs/C"], success=True, mergelist=["dev-libs/C-1"]
            ),
            ResolverPlaygroundTestCase(
                ["dev-libs/D"], success=True, mergelist=["dev-libs/D-1"]
            ),
            ResolverPlaygroundTestCase(
                ["dev-libs/E"], options={"--autounmask": "n"}, success=False
            ),
        )

        playground = ResolverPlayground(
            ebuilds=ebuilds,
            repo_configs=repo_configs,
            profile=profile,
            user_config=user_config,
        )
        try:
            for test_case in test_cases:
                playground.run_TestCase(test_case)
                self.assertEqual(test_case.test_success, True, test_case.fail_msg)
        finally:
            playground.cleanup()

    def testManifest(self):

        distfiles = {
            "B-2.tar.bz2": b"binary\0content",
            "C-2.zip": b"binary\0content",
            "C-2.tar.bz2": b"binary\0content",
        }

        ebuilds = {
            "dev-libs/A-1::old_repo": {},
            "dev-libs/A-2::new_repo": {},
            "dev-libs/B-2::new_repo": {"SRC_URI": "B-2.tar.bz2"},
            "dev-libs/C-2::new_repo": {"SRC_URI": "C-2.zip C-2.tar.bz2"},
        }

        repo_configs = {
            "new_repo": {
                "layout.conf": (
                    "profile-formats = pms",
                    "thin-manifests = true",
                    "manifest-hashes = SHA256 SHA512 WHIRLPOOL",
                    "manifest-required-hashes = SHA512",
                    "# use implicit masters",
                ),
            }
        }

        test_cases = (
            ResolverPlaygroundTestCase(
                ["=dev-libs/A-1"], mergelist=["dev-libs/A-1"], success=True
            ),
            ResolverPlaygroundTestCase(
                ["=dev-libs/A-2"], mergelist=["dev-libs/A-2"], success=True
            ),
        )

        playground = ResolverPlayground(
            ebuilds=ebuilds, repo_configs=repo_configs, distfiles=distfiles
        )
        settings = playground.settings

        new_repo_config = settings.repositories["new_repo"]
        old_repo_config = settings.repositories["old_repo"]
        self.assertTrue(
            len(new_repo_config.masters) > 0, "new_repo has no default master"
        )
        self.assertEqual(
            new_repo_config.masters[0].location,
            playground.settings.repositories["test_repo"].location,
            "new_repo default master is not test_repo",
        )
        self.assertEqual(
            new_repo_config.thin_manifest, True, "new_repo_config.thin_manifest != True"
        )

        new_manifest_file = os.path.join(
            new_repo_config.location, "dev-libs", "A", "Manifest"
        )
        self.assertNotExists(new_manifest_file)

        new_manifest_file = os.path.join(
            new_repo_config.location, "dev-libs", "B", "Manifest"
        )
        f = open(new_manifest_file)
        self.assertEqual(len(list(f)), 1)
        f.close()

        new_manifest_file = os.path.join(
            new_repo_config.location, "dev-libs", "C", "Manifest"
        )
        f = open(new_manifest_file)
        self.assertEqual(len(list(f)), 2)
        f.close()

        old_manifest_file = os.path.join(
            old_repo_config.location, "dev-libs", "A", "Manifest"
        )
        f = open(old_manifest_file)
        self.assertEqual(len(list(f)), 1)
        f.close()

        try:
            for test_case in test_cases:
                playground.run_TestCase(test_case)
                self.assertEqual(test_case.test_success, True, test_case.fail_msg)
        finally:
            playground.cleanup()

    def testSetCpv(self):
        """
        Test the clone via constructor.
        """

        ebuilds = {
            "dev-libs/A-1": {"IUSE": "static-libs"},
            "dev-libs/B-1": {"IUSE": "static-libs"},
        }

        env_files = {"A": ('USE="static-libs"',)}

        package_env = ("dev-libs/A A",)

        eprefix = normalize_path(tempfile.mkdtemp())
        playground = None
        try:
            user_config_dir = os.path.join(eprefix, USER_CONFIG_PATH)
            os.makedirs(user_config_dir)

            with open(
                os.path.join(user_config_dir, "package.env"),
                mode="w",
                encoding=_encodings["content"],
            ) as f:
                for line in package_env:
                    f.write(line + "\n")

            env_dir = os.path.join(user_config_dir, "env")
            os.makedirs(env_dir)
            for k, v in env_files.items():
                with open(
                    os.path.join(env_dir, k), mode="w", encoding=_encodings["content"]
                ) as f:
                    for line in v:
                        f.write(line + "\n")

            playground = ResolverPlayground(eprefix=eprefix, ebuilds=ebuilds)
            settings = config(clone=playground.settings)

            result = playground.run(["=dev-libs/A-1"])
            pkg, existing_node = result.depgraph._select_package(
                playground.eroot, Atom("=dev-libs/A-1")
            )
            settings.setcpv(pkg)
            self.assertTrue("static-libs" in settings["PORTAGE_USE"].split())

            # Test bug #522362, where a USE=static-libs package.env
            # setting leaked from one setcpv call to the next.
            pkg, existing_node = result.depgraph._select_package(
                playground.eroot, Atom("=dev-libs/B-1")
            )
            settings.setcpv(pkg)
            self.assertTrue("static-libs" not in settings["PORTAGE_USE"].split())

        finally:
            if playground is None:
                shutil.rmtree(eprefix)
            else:
                playground.cleanup()



# Copyright 1999-2009 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

import portage


class UninstallFailure(portage.exception.PortageException):
    """
    An instance of this class is raised by unmerge() when
    an uninstallation fails.
    """

    status = 1

    def __init__(self, *pargs):
        portage.exception.PortageException.__init__(self, pargs)
        if pargs:
            self.status = pargs[0]



# Copyright 1999-2020 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

try:
    import threading
except ImportError:
    import dummy_threading as threading

from portage.util.futures import asyncio
from portage.util._async.SchedulerInterface import SchedulerInterface
from portage.util._eventloop.global_event_loop import global_event_loop

from _emerge.getloadavg import getloadavg


class PollScheduler:

    # max time between loadavg checks (milliseconds)
    _loadavg_latency = None

    def __init__(self, main=False, event_loop=None):
        """
        @param main: If True then use global_event_loop(), otherwise use
                a local EventLoop instance (default is False, for safe use in
                a non-main thread)
        @type main: bool
        """
        self._term_rlock = threading.RLock()
        self._terminated = threading.Event()
        self._terminated_tasks = False
        self._term_check_handle = None
        self._max_jobs = 1
        self._max_load = None
        self._scheduling = False
        self._background = False
        if event_loop is not None:
            self._event_loop = event_loop
        elif main:
            self._event_loop = global_event_loop()
        else:
            self._event_loop = asyncio._safe_loop()
        self._sched_iface = SchedulerInterface(
            self._event_loop, is_background=self._is_background
        )

    def _is_background(self):
        return self._background

    def _cleanup(self):
        """
        Cleanup any callbacks that have been registered with the global
        event loop.
        """
        # The self._term_check_handle attribute requires locking
        # since it's modified by the thread safe terminate method.
        with self._term_rlock:
            if self._term_check_handle not in (None, False):
                self._term_check_handle.cancel()
            # This prevents the terminate method from scheduling
            # any more callbacks (since _cleanup must eliminate all
            # callbacks in order to ensure complete cleanup).
            self._term_check_handle = False

    def terminate(self):
        """
        Schedules asynchronous, graceful termination of the scheduler
        at the earliest opportunity.

        This method is thread-safe (and safe for signal handlers).
        """
        with self._term_rlock:
            if self._term_check_handle is None:
                self._terminated.set()
                self._term_check_handle = self._event_loop.call_soon_threadsafe(
                    self._termination_check, True
                )

    def _termination_check(self, retry=False):
        """
        Calls _terminate_tasks() if appropriate. It's guaranteed not to
        call it while _schedule_tasks() is being called. This method must
        only be called via the event loop thread.

        @param retry: If True then reschedule if scheduling state prevents
                immediate termination.
        @type retry: bool
        """
        if self._terminated.is_set() and not self._terminated_tasks:
            if not self._scheduling:
                self._scheduling = True
                try:
                    self._terminated_tasks = True
                    self._terminate_tasks()
                finally:
                    self._scheduling = False

            elif retry:
                with self._term_rlock:
                    self._term_check_handle = self._event_loop.call_soon(
                        self._termination_check, True
                    )

    def _terminate_tasks(self):
        """
        Send signals to terminate all tasks. This is called once
        from _keep_scheduling() or _is_work_scheduled() in the event
        dispatching thread. It will not be called while the _schedule_tasks()
        implementation is running, in order to avoid potential
        interference. All tasks should be cleaned up at the earliest
        opportunity, but not necessarily before this method returns.
        Typically, this method will send kill signals and return without
        waiting for exit status. This allows basic cleanup to occur, such as
        flushing of buffered output to logs.
        """
        raise NotImplementedError()

    def _keep_scheduling(self):
        """
        @rtype: bool
        @return: True if there may be remaining tasks to schedule,
                False otherwise.
        """
        return False

    def _schedule_tasks(self):
        """
        This is called from inside the _schedule() method, which
        guarantees the following:

        1) It will not be called recursively.
        2) _terminate_tasks() will not be called while it is running.
        3) The state of the boolean _terminated_tasks variable will
           not change while it is running.

        Unless this method is used to perform user interface updates,
        or something like that, the first thing it should do is check
        the state of _terminated_tasks and if that is True then it
        should return immediately (since there's no need to
        schedule anything after _terminate_tasks() has been called).
        """
        pass

    def _schedule(self):
        """
        Calls _schedule_tasks() and automatically returns early from
        any recursive calls to this method that the _schedule_tasks()
        call might trigger. This makes _schedule() safe to call from
        inside exit listeners. This method always returns True, so that
        it may be scheduled continuously via EventLoop.timeout_add().
        """
        if self._scheduling:
            return True
        self._scheduling = True
        try:
            self._schedule_tasks()
        finally:
            self._scheduling = False
        return True

    def _is_work_scheduled(self):
        return bool(self._running_job_count())

    def _running_job_count(self):
        raise NotImplementedError(self)

    def _can_add_job(self):
        if self._terminated_tasks:
            return False

        max_jobs = self._max_jobs
        max_load = self._max_load

        if self._max_jobs is not True and self._running_job_count() >= self._max_jobs:
            return False

        if (
            max_load is not None
            and (max_jobs is True or max_jobs > 1)
            and self._running_job_count() >= 1
        ):
            try:
                avg1, avg5, avg15 = getloadavg()
            except OSError:
                return False

            if avg1 >= max_load:
                return False

        return True



# Copyright 2006-2009 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

from portage import os
from portage.tests import TestCase
from portage.env.config import PortageModulesFile
from tempfile import mkstemp


class PortageModulesFileTestCase(TestCase):

    keys = ["foo.bar", "baz", "bob", "extra_key"]
    invalid_keys = ["", ""]
    modules = ["spanky", "zmedico", "antarus", "ricer", "5", "6"]

    def setUp(self):
        self.items = {}
        for k, v in zip(self.keys + self.invalid_keys, self.modules):
            self.items[k] = v

    def testPortageModulesFile(self):
        self.BuildFile()
        f = PortageModulesFile(self.fname)
        f.load()
        for k in self.keys:
            self.assertEqual(f[k], self.items[k])
        for ik in self.invalid_keys:
            self.assertEqual(False, ik in f)
        self.NukeFile()

    def BuildFile(self):
        fd, self.fname = mkstemp()
        f = os.fdopen(fd, "w")
        for k, v in self.items.items():
            f.write("{}={}\n".format(k, v))
        f.close()

    def NukeFile(self):
        os.unlink(self.fname)



# Copyright 2014-2015 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

import errno
import json
import os

from portage import _encodings
from portage.util import atomic_ofstream
from portage.versions import cpv_getkey


class VdbMetadataDelta:

    _format_version = "1"

    def __init__(self, vardb):
        self._vardb = vardb

    def initialize(self, timestamp):
        with atomic_ofstream(
            self._vardb._cache_delta_filename,
            "w",
            encoding=_encodings["repo.content"],
            errors="strict",
        ) as f:
            json.dump(
                {"version": self._format_version, "timestamp": timestamp},
                f,
                ensure_ascii=False,
            )

    def load(self):

        if not os.path.exists(self._vardb._aux_cache_filename):
            # If the primary cache doesn't exist yet, then
            # we can't record a delta against it.
            return None

        try:
            with open(
                self._vardb._cache_delta_filename,
                encoding=_encodings["repo.content"],
                errors="strict",
            ) as f:
                cache_obj = json.load(f)
        except OSError as e:
            if e.errno not in (errno.ENOENT, errno.ESTALE):
                raise
        except (SystemExit, KeyboardInterrupt):
            raise
        except Exception:
            # Corrupt, or not json format.
            pass
        else:
            try:
                version = cache_obj["version"]
            except KeyError:
                pass
            else:
                # Verify that the format version is compatible,
                # since a newer version of portage may have
                # written an incompatible file.
                if version == self._format_version:
                    try:
                        deltas = cache_obj["deltas"]
                    except KeyError:
                        cache_obj["deltas"] = deltas = []

                    if isinstance(deltas, list):
                        return cache_obj

        return None

    def loadRace(self):
        """
        This calls self.load() and validates the timestamp
        against the currently loaded self._vardb._aux_cache. If a
        concurrent update causes the timestamps to be inconsistent,
        then it reloads the caches and tries one more time before
        it aborts. In practice, the race is very unlikely, so
        this will usually succeed on the first try.
        """

        tries = 2
        while tries:
            tries -= 1
            cache_delta = self.load()
            if cache_delta is not None and cache_delta.get(
                "timestamp"
            ) != self._vardb._aux_cache.get("timestamp", False):
                self._vardb._aux_cache_obj = None
            else:
                return cache_delta

        return None

    def recordEvent(self, event, cpv, slot, counter):

        self._vardb.lock()
        try:
            deltas_obj = self.load()

            if deltas_obj is None:
                # We can't record meaningful deltas without
                # a pre-existing state.
                return

            delta_node = {
                "event": event,
                "package": cpv.cp,
                "version": cpv.version,
                "slot": slot,
                "counter": "%s" % counter,
            }

            deltas_obj["deltas"].append(delta_node)

            # Eliminate earlier nodes cancelled out by later nodes
            # that have identical package and slot attributes.
            filtered_list = []
            slot_keys = set()
            version_keys = set()
            for delta_node in reversed(deltas_obj["deltas"]):
                slot_key = (delta_node["package"], delta_node["slot"])
                version_key = (delta_node["package"], delta_node["version"])
                if not (slot_key in slot_keys or version_key in version_keys):
                    filtered_list.append(delta_node)
                    slot_keys.add(slot_key)
                    version_keys.add(version_key)

            filtered_list.reverse()
            deltas_obj["deltas"] = filtered_list

            f = atomic_ofstream(
                self._vardb._cache_delta_filename,
                mode="w",
                encoding=_encodings["repo.content"],
            )
            json.dump(deltas_obj, f, ensure_ascii=False)
            f.close()

        finally:
            self._vardb.unlock()

    def applyDelta(self, data):
        packages = self._vardb._aux_cache["packages"]
        deltas = {}
        for delta in data["deltas"]:
            cpv = delta["package"] + "-" + delta["version"]
            deltas[cpv] = delta
            event = delta["event"]
            if event == "add":
                # Use aux_get to populate the cache
                # for this cpv.
                if cpv not in packages:
                    try:
                        self._vardb.aux_get(cpv, ["DESCRIPTION"])
                    except KeyError:
                        pass
            elif event == "remove":
                packages.pop(cpv, None)

        if deltas:
            # Delete removed or replaced versions from affected slots
            for cached_cpv, (mtime, metadata) in list(packages.items()):
                if cached_cpv in deltas:
                    continue

                removed = False
                for cpv, delta in deltas.items():
                    if (
                        cached_cpv.startswith(delta["package"])
                        and metadata.get("SLOT") == delta["slot"]
                        and cpv_getkey(cached_cpv) == delta["package"]
                    ):
                        removed = True
                        break

                if removed:
                    del packages[cached_cpv]
                    del deltas[cpv]
                    if not deltas:
                        break



# Copyright 2020 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

from portage.const import SUPPORTED_GENTOO_BINPKG_FORMATS
from portage.tests import TestCase
from portage.tests.resolver.ResolverPlayground import (
    ResolverPlayground,
    ResolverPlaygroundTestCase,
)
from portage.output import colorize


class SlotOperatorBdependTestCase(TestCase):
    def testSlotOperatorBdepend(self):
        """
        Test regular dev-lang/go upgrade, with rebuild of packages
        that have dev-lang/go:= in BDEPEND.
        """

        ebuilds = {
            "app-emulation/buildah-1.16.1": {
                "EAPI": "7",
                "BDEPEND": "dev-lang/go:=",
            },
            "app-emulation/libpod-2.1.0": {
                "EAPI": "7",
                "BDEPEND": "dev-lang/go:=",
            },
            "dev-lang/go-1.15.5": {"EAPI": "7", "SLOT": "0/1.15.5"},
            "dev-lang/go-1.14.12": {"EAPI": "7", "SLOT": "0/1.14.12"},
        }

        binpkgs = {
            "app-emulation/buildah-1.16.1": {
                "EAPI": "7",
                "BDEPEND": "dev-lang/go:0/1.14.12=",
            },
            "app-emulation/libpod-2.1.0": {
                "EAPI": "7",
                "BDEPEND": "dev-lang/go:0/1.14.12=",
            },
            "dev-lang/go-1.14.12": {"EAPI": "7", "SLOT": "0/1.14.12"},
        }

        installed = {
            "app-emulation/buildah-1.16.1": {
                "EAPI": "7",
                "BDEPEND": "dev-lang/go:0/1.14.12=",
            },
            "app-emulation/libpod-2.1.0": {
                "EAPI": "7",
                "BDEPEND": "dev-lang/go:0/1.14.12=",
            },
            "dev-lang/go-1.14.12": {"EAPI": "7", "SLOT": "0/1.14.12"},
        }

        world = ["app-emulation/buildah", "app-emulation/libpod"]

        test_cases = (
            # Test rebuild triggered by slot operator := dependency in BDEPEND.
            ResolverPlaygroundTestCase(
                ["@world"],
                options={
                    "--update": True,
                    "--deep": True,
                },
                success=True,
                mergelist=[
                    "dev-lang/go-1.15.5",
                    "app-emulation/buildah-1.16.1",
                    "app-emulation/libpod-2.1.0",
                ],
            ),
            # Test the above case with --usepkg --with-bdeps=y. It should not use the
            # binary packages because rebuild is needed.
            ResolverPlaygroundTestCase(
                ["@world"],
                options={
                    "--usepkg": True,
                    "--with-bdeps": "y",
                    "--update": True,
                    "--deep": True,
                },
                success=True,
                mergelist=[
                    "dev-lang/go-1.15.5",
                    "app-emulation/buildah-1.16.1",
                    "app-emulation/libpod-2.1.0",
                ],
            ),
        )

        for binpkg_format in SUPPORTED_GENTOO_BINPKG_FORMATS:
            with self.subTest(binpkg_format=binpkg_format):
                print(colorize("HILITE", binpkg_format), end=" ... ")
                playground = ResolverPlayground(
                    ebuilds=ebuilds,
                    binpkgs=binpkgs,
                    installed=installed,
                    world=world,
                    debug=False,
                    user_config={
                        "make.conf": ('BINPKG_FORMAT="%s"' % binpkg_format,),
                    },
                )
                try:
                    for test_case in test_cases:
                        playground.run_TestCase(test_case)
                        self.assertEqual(
                            test_case.test_success, True, test_case.fail_msg
                        )
                finally:
                    playground.debug = False
                    playground.cleanup()

    def testSlotOperatorBdependAfterBreakage(self):
        """
        Test rebuild of packages that have dev-lang/go:= in BDEPEND,
        after the built slot operator deps have already been broken
        by an earlier dev-lang/go upgrade.
        """

        ebuilds = {
            "app-emulation/buildah-1.16.1": {
                "EAPI": "7",
                "BDEPEND": "dev-lang/go:=",
            },
            "app-emulation/libpod-2.1.0": {
                "EAPI": "7",
                "BDEPEND": "dev-lang/go:=",
            },
            "dev-lang/go-1.15.5": {"EAPI": "7", "SLOT": "0/1.15.5"},
            "dev-lang/go-1.14.12": {"EAPI": "7", "SLOT": "0/1.14.12"},
        }

        binpkgs = {
            "app-emulation/buildah-1.16.1": {
                "EAPI": "7",
                "BDEPEND": "dev-lang/go:0/1.14.12=",
            },
            "app-emulation/libpod-2.1.0": {
                "EAPI": "7",
                "BDEPEND": "dev-lang/go:0/1.14.12=",
            },
            "dev-lang/go-1.14.12": {"EAPI": "7", "SLOT": "0/1.14.12"},
            "dev-lang/go-1.15.5": {"EAPI": "7", "SLOT": "0/1.15.5"},
        }

        installed = {
            "app-emulation/buildah-1.16.1": {
                "EAPI": "7",
                "BDEPEND": "dev-lang/go:0/1.14.12=",
            },
            "app-emulation/libpod-2.1.0": {
                "EAPI": "7",
                "BDEPEND": "dev-lang/go:0/1.14.12=",
            },
            "dev-lang/go-1.15.5": {"EAPI": "7", "SLOT": "0/1.15.5"},
        }

        world = ["app-emulation/buildah", "app-emulation/libpod"]

        test_cases = (
            # Test rebuild triggered by slot operator := dependency in BDEPEND.
            ResolverPlaygroundTestCase(
                ["@world"],
                options={
                    "--update": True,
                    "--deep": True,
                },
                success=True,
                mergelist=[
                    "app-emulation/buildah-1.16.1",
                    "app-emulation/libpod-2.1.0",
                ],
            ),
            # Test the above case with --usepkg --with-bdeps=y. It should not use the
            # binary packages because rebuild is needed.
            ResolverPlaygroundTestCase(
                ["@world"],
                options={
                    "--usepkg": True,
                    "--with-bdeps": "y",
                    "--update": True,
                    "--deep": True,
                },
                success=True,
                mergelist=[
                    "app-emulation/buildah-1.16.1",
                    "app-emulation/libpod-2.1.0",
                ],
            ),
        )

        for binpkg_format in SUPPORTED_GENTOO_BINPKG_FORMATS:
            with self.subTest(binpkg_format=binpkg_format):
                print(colorize("HILITE", binpkg_format), end=" ... ")
                playground = ResolverPlayground(
                    ebuilds=ebuilds,
                    binpkgs=binpkgs,
                    installed=installed,
                    world=world,
                    debug=False,
                    user_config={
                        "make.conf": ('BINPKG_FORMAT="%s"' % binpkg_format,),
                    },
                )
                try:
                    for test_case in test_cases:
                        playground.run_TestCase(test_case)
                        self.assertEqual(
                            test_case.test_success, True, test_case.fail_msg
                        )
                finally:
                    playground.debug = False
                    playground.cleanup()



# Copyright 2011 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

from portage.tests import TestCase
from portage.tests.resolver.ResolverPlayground import (
    ResolverPlayground,
    ResolverPlaygroundTestCase,
)


class VirtualSlotResolverTestCase(TestCase):
    def testLicenseMaskedVirtualSlotUpdate(self):

        ebuilds = {
            "dev-java/oracle-jdk-bin-1.7.0": {"SLOT": "1.7", "LICENSE": "TEST"},
            "dev-java/sun-jdk-1.6.0": {"SLOT": "1.6", "LICENSE": "TEST"},
            "dev-java/icedtea-6.1.10.3": {"SLOT": "6"},
            "app-misc/java-app-1": {"RDEPEND": ">=virtual/jdk-1.6.0"},
            "virtual/jdk-1.6.0": {
                "SLOT": "1.6",
                "RDEPEND": "|| ( =dev-java/icedtea-6* =dev-java/sun-jdk-1.6.0* )",
            },
            "virtual/jdk-1.7.0": {
                "SLOT": "1.7",
                "RDEPEND": "|| ( =dev-java/oracle-jdk-bin-1.7.0* )",
            },
        }

        installed = {
            "app-misc/java-app-1": {"RDEPEND": ">=virtual/jdk-1.6.0"},
            "dev-java/icedtea-6.1.10.3": {"SLOT": "6"},
            "virtual/jdk-1.6.0": {
                "SLOT": "1.6",
                "RDEPEND": "|| ( =dev-java/icedtea-6* =dev-java/sun-jdk-1.6.0* )",
            },
        }

        world = ("app-misc/java-app",)

        test_cases = (
            # Bug #382557 - Don't pull in the virtual/jdk-1.7.0 slot update
            # since its dependencies can only be satisfied by a package that
            # is masked by license.
            ResolverPlaygroundTestCase(
                ["@world"],
                options={"--update": True, "--deep": True},
                success=True,
                mergelist=[],
            ),
        )

        playground = ResolverPlayground(
            ebuilds=ebuilds, installed=installed, world=world
        )
        try:
            for test_case in test_cases:
                playground.run_TestCase(test_case)
                self.assertEqual(test_case.test_success, True, test_case.fail_msg)
        finally:
            playground.cleanup()

    def testVirtualSlotUpdate(self):

        ebuilds = {
            "dev-java/oracle-jdk-bin-1.7.0": {"SLOT": "1.7", "LICENSE": "TEST"},
            "dev-java/sun-jdk-1.6.0": {"SLOT": "1.6", "LICENSE": "TEST"},
            "dev-java/icedtea-6.1.10.3": {"SLOT": "6"},
            "dev-java/icedtea-7": {"SLOT": "7"},
            "app-misc/java-app-1": {"RDEPEND": ">=virtual/jdk-1.6.0"},
            "virtual/jdk-1.6.0": {
                "SLOT": "1.6",
                "RDEPEND": "|| ( =dev-java/icedtea-6* =dev-java/sun-jdk-1.6.0* )",
            },
            "virtual/jdk-1.7.0": {
                "SLOT": "1.7",
                "RDEPEND": "|| ( =dev-java/icedtea-7* =dev-java/oracle-jdk-bin-1.7.0* )",
            },
        }

        installed = {
            "app-misc/java-app-1": {"RDEPEND": ">=virtual/jdk-1.6.0"},
            "dev-java/icedtea-6.1.10.3": {"SLOT": "6"},
            "virtual/jdk-1.6.0": {
                "SLOT": "1.6",
                "RDEPEND": "|| ( =dev-java/icedtea-6* =dev-java/sun-jdk-1.6.0* )",
            },
        }

        world = ("app-misc/java-app",)

        test_cases = (
            # Pull in the virtual/jdk-1.7.0 slot update since its dependencies
            # can only be satisfied by an unmasked package.
            ResolverPlaygroundTestCase(
                ["@world"],
                options={"--update": True, "--deep": True},
                success=True,
                mergelist=["dev-java/icedtea-7", "virtual/jdk-1.7.0"],
            ),
            # Bug #275945 - Don't pull in the virtual/jdk-1.7.0 slot update
            # unless --update is enabled.
            ResolverPlaygroundTestCase(
                ["@world"],
                options={"--selective": True, "--deep": True},
                success=True,
                mergelist=[],
            ),
        )

        playground = ResolverPlayground(
            ebuilds=ebuilds, installed=installed, world=world
        )
        try:
            for test_case in test_cases:
                playground.run_TestCase(test_case)
                self.assertEqual(test_case.test_success, True, test_case.fail_msg)
        finally:
            playground.cleanup()

    def testVirtualSubslotUpdate(self):

        ebuilds = {
            "virtual/pypy-2.3.1": {
                "EAPI": "5",
                "SLOT": "0/2.3",
                "RDEPEND": "|| ( >=dev-python/pypy-2.3.1:0/2.3 >=dev-python/pypy-bin-2.3.1:0/2.3 ) ",
            },
            "virtual/pypy-2.4.0": {
                "EAPI": "5",
                "SLOT": "0/2.4",
                "RDEPEND": "|| ( >=dev-python/pypy-2.4.0:0/2.4 >=dev-python/pypy-bin-2.4.0:0/2.4 ) ",
            },
            "dev-python/pypy-2.3.1": {"EAPI": "5", "SLOT": "0/2.3"},
            "dev-python/pypy-2.4.0": {"EAPI": "5", "SLOT": "0/2.4"},
            "dev-python/pygments-1.6_p20140324-r1": {
                "EAPI": "5",
                "DEPEND": "virtual/pypy:0=",
            },
        }

        installed = {
            "virtual/pypy-2.3.1": {
                "EAPI": "5",
                "SLOT": "0/2.3",
                "RDEPEND": "|| ( >=dev-python/pypy-2.3.1:0/2.3 >=dev-python/pypy-bin-2.3.1:0/2.3 ) ",
            },
            "dev-python/pypy-2.3.1": {"EAPI": "5", "SLOT": "0/2.3"},
            "dev-python/pygments-1.6_p20140324-r1": {
                "EAPI": "5",
                "DEPEND": "virtual/pypy:0/2.3=",
                "RDEPEND": "virtual/pypy:0/2.3=",
            },
        }

        world = ["dev-python/pygments"]

        test_cases = (
            # bug 526160 - test for missed pypy sub-slot update
            ResolverPlaygroundTestCase(
                ["@world"],
                options={"--update": True, "--deep": True, "--dynamic-deps": "y"},
                success=True,
                mergelist=[
                    "dev-python/pypy-2.4.0",
                    "virtual/pypy-2.4.0",
                    "dev-python/pygments-1.6_p20140324-r1",
                ],
            ),
            # Repeat above test, but with --dynamic-deps disabled.
            ResolverPlaygroundTestCase(
                ["@world"],
                options={"--update": True, "--deep": True, "--dynamic-deps": "n"},
                success=True,
                mergelist=[
                    "dev-python/pypy-2.4.0",
                    "virtual/pypy-2.4.0",
                    "dev-python/pygments-1.6_p20140324-r1",
                ],
            ),
        )

        playground = ResolverPlayground(
            debug=False, ebuilds=ebuilds, installed=installed, world=world
        )
        try:
            for test_case in test_cases:
                playground.run_TestCase(test_case)
                self.assertEqual(test_case.test_success, True, test_case.fail_msg)
        finally:
            playground.cleanup()

    def testVirtualSlotDepclean(self):

        ebuilds = {
            "dev-java/oracle-jdk-bin-1.7.0": {"SLOT": "1.7", "LICENSE": "TEST"},
            "dev-java/sun-jdk-1.6.0": {"SLOT": "1.6", "LICENSE": "TEST"},
            "dev-java/icedtea-6.1.10.3": {"SLOT": "6"},
            "dev-java/icedtea-7": {"SLOT": "7"},
            "app-misc/java-app-1": {"RDEPEND": ">=virtual/jdk-1.6.0"},
            "virtual/jdk-1.6.0": {
                "SLOT": "1.6",
                "RDEPEND": "|| ( =dev-java/icedtea-6* =dev-java/sun-jdk-1.6.0* )",
            },
            "virtual/jdk-1.7.0": {
                "SLOT": "1.7",
                "RDEPEND": "|| ( =dev-java/icedtea-7* =dev-java/oracle-jdk-bin-1.7.0* )",
            },
        }

        installed = {
            "app-misc/java-app-1": {"RDEPEND": ">=virtual/jdk-1.6.0"},
            "dev-java/icedtea-6.1.10.3": {"SLOT": "6"},
            "dev-java/icedtea-7": {"SLOT": "7"},
            "virtual/jdk-1.6.0": {
                "SLOT": "1.6",
                "RDEPEND": "|| ( =dev-java/icedtea-6* =dev-java/sun-jdk-1.6.0* )",
            },
            "virtual/jdk-1.7.0": {
                "SLOT": "1.7",
                "RDEPEND": "|| ( =dev-java/icedtea-7* =dev-java/oracle-jdk-bin-1.7.0* )",
            },
        }

        world = (
            "virtual/jdk:1.6",
            "app-misc/java-app",
        )

        test_cases = (
            # Make sure that depclean doesn't remove a new slot even though
            # it is redundant in the sense that the older slot will satisfy
            # all dependencies.
            ResolverPlaygroundTestCase(
                [], options={"--depclean": True}, success=True, cleanlist=[]
            ),
            # Prune redundant lower slots, even if they are in world.
            ResolverPlaygroundTestCase(
                [],
                options={"--prune": True},
                success=True,
                cleanlist=["virtual/jdk-1.6.0", "dev-java/icedtea-6.1.10.3"],
            ),
        )

        playground = ResolverPlayground(
            ebuilds=ebuilds, installed=installed, world=world
        )
        try:
            for test_case in test_cases:
                playground.run_TestCase(test_case)
                self.assertEqual(test_case.test_success, True, test_case.fail_msg)
        finally:
            playground.cleanup()



# Copyright 2020-2021 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

from portage.tests import TestCase
from portage.tests.resolver.ResolverPlayground import ResolverPlayground
from portage.util.futures import asyncio
from portage.util.futures.executor.fork import ForkExecutor


class AuxdbTestCase(TestCase):
    def test_anydbm(self):
        try:
            from portage.cache.anydbm import database
        except ImportError:
            self.skipTest("dbm import failed")
        self._test_mod("portage.cache.anydbm.database", multiproc=False)

    def test_flat_hash_md5(self):
        self._test_mod("portage.cache.flat_hash.md5_database")

    def test_volatile(self):
        self._test_mod("portage.cache.volatile.database", multiproc=False)

    def test_sqite(self):
        try:
            import sqlite3
        except ImportError:
            self.skipTest("sqlite3 import failed")
        self._test_mod("portage.cache.sqlite.database")

    def _test_mod(self, auxdbmodule, multiproc=True):
        ebuilds = {
            "cat/A-1": {
                "EAPI": "7",
                "MISC_CONTENT": "inherit foo",
            },
            "cat/B-1": {
                "EAPI": "7",
                "MISC_CONTENT": "inherit foo",
            },
        }

        ebuild_inherited = frozenset(["bar", "foo"])
        eclass_defined_phases = "prepare"
        eclass_depend = "bar/foo"

        eclasses = {
            "foo": ("inherit bar",),
            "bar": (
                "EXPORT_FUNCTIONS src_prepare",
                'DEPEND="{}"'.format(eclass_depend),
                "bar_src_prepare() { default; }",
            ),
        }

        playground = ResolverPlayground(
            ebuilds=ebuilds,
            eclasses=eclasses,
            user_config={"modules": ("portdbapi.auxdbmodule = %s" % auxdbmodule,)},
        )

        portdb = playground.trees[playground.eroot]["porttree"].dbapi

        def test_func():
            loop = asyncio._wrap_loop()
            return loop.run_until_complete(
                self._test_mod_async(
                    ebuilds,
                    ebuild_inherited,
                    eclass_defined_phases,
                    eclass_depend,
                    portdb,
                )
            )

        self.assertTrue(test_func())

        loop = asyncio._wrap_loop()
        self.assertTrue(
            loop.run_until_complete(loop.run_in_executor(ForkExecutor(), test_func))
        )

        auxdb = portdb.auxdb[portdb.getRepositoryPath("test_repo")]
        cpv = next(iter(ebuilds))

        def modify_auxdb():
            metadata = auxdb[cpv]
            metadata["RESTRICT"] = "test"
            try:
                del metadata["_eclasses_"]
            except KeyError:
                pass
            auxdb[cpv] = metadata

        if multiproc:
            loop.run_until_complete(loop.run_in_executor(ForkExecutor(), modify_auxdb))
        else:
            modify_auxdb()

        self.assertEqual(auxdb[cpv]["RESTRICT"], "test")

    async def _test_mod_async(
        self, ebuilds, ebuild_inherited, eclass_defined_phases, eclass_depend, portdb
    ):

        for cpv, metadata in ebuilds.items():
            defined_phases, depend, eapi, inherited = await portdb.async_aux_get(
                cpv, ["DEFINED_PHASES", "DEPEND", "EAPI", "INHERITED"]
            )
            self.assertEqual(defined_phases, eclass_defined_phases)
            self.assertEqual(depend, eclass_depend)
            self.assertEqual(eapi, metadata["EAPI"])
            self.assertEqual(frozenset(inherited.split()), ebuild_inherited)

        return True



# Copyright 1999-2009 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

import time


class ProgressHandler:
    def __init__(self):
        self.curval = 0
        self.maxval = 0
        self._last_update = 0
        self.min_latency = 0.2

    def onProgress(self, maxval, curval):
        self.maxval = maxval
        self.curval = curval
        cur_time = time.time()
        if cur_time - self._last_update >= self.min_latency:
            self._last_update = cur_time
            self.display()

    def display(self):
        raise NotImplementedError(self)



# Copyright 2015-2020 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import pickle
import traceback

from portage import os
from portage.util._async.ForkProcess import ForkProcess
from _emerge.PipeReader import PipeReader


class AsyncFunction(ForkProcess):
    """
    Execute a function call in a fork, and retrieve the function
    return value via pickling/unpickling, accessible as the
    "result" attribute after the forked process has exited.
    """

    # NOTE: This class overrides the meaning of the SpawnProcess 'args'
    # attribute, and uses it to hold the positional arguments for the
    # 'target' function.
    __slots__ = (
        "kwargs",
        "result",
        "target",
        "_async_func_reader",
        "_async_func_reader_pw",
    )

    def _start(self):
        pr, pw = os.pipe()
        self.fd_pipes = {} if self.fd_pipes is None else self.fd_pipes
        self.fd_pipes[pw] = pw
        self._async_func_reader_pw = pw
        self._async_func_reader = PipeReader(
            input_files={"input": pr}, scheduler=self.scheduler
        )
        self._async_func_reader.addExitListener(self._async_func_reader_exit)
        self._async_func_reader.start()
        ForkProcess._start(self)
        os.close(pw)

    def _run(self):
        try:
            result = self.target(*(self.args or []), **(self.kwargs or {}))
            os.write(self._async_func_reader_pw, pickle.dumps(result))
        except Exception:
            traceback.print_exc()
            return 1

        return os.EX_OK

    def _async_waitpid(self):
        # Ignore this event, since we want to ensure that we exit
        # only after _async_func_reader_exit has reached EOF.
        if self._async_func_reader is None:
            ForkProcess._async_waitpid(self)

    def _async_func_reader_exit(self, pipe_reader):
        try:
            self.result = pickle.loads(pipe_reader.getvalue())
        except Exception:
            # The child process will have printed a traceback in this case,
            # and returned an unsuccessful returncode.
            pass
        self._async_func_reader = None
        if self.returncode is None:
            self._async_waitpid()
        else:
            self._unregister()
            self._async_wait()

    def _unregister(self):
        ForkProcess._unregister(self)

        pipe_reader = self._async_func_reader
        if pipe_reader is not None:
            self._async_func_reader = None
            pipe_reader.removeExitListener(self._async_func_reader_exit)
            pipe_reader.cancel()



# Copyright 1999-2011 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

import copy
from portage.cache import template


class database(template.database):

    autocommits = True
    serialize_eclasses = False
    store_eclass_paths = False

    def __init__(self, *args, **config):
        config.pop("gid", None)
        config.pop("perms", None)
        super().__init__(*args, **config)
        self._data = {}
        self._delitem = self._data.__delitem__

    def _setitem(self, name, values):
        self._data[name] = copy.deepcopy(values)

    def __getitem__(self, cpv):
        return copy.deepcopy(self._data[cpv])

    def __iter__(self):
        return iter(self._data)

    def __contains__(self, key):
        return key in self._data



# Copyright 2010 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

from portage.package.ebuild._ipc.IpcCommand import IpcCommand


class ExitCommand(IpcCommand):

    __slots__ = (
        "exitcode",
        "reply_hook",
    )

    def __init__(self):
        IpcCommand.__init__(self)
        self.reply_hook = None
        self.exitcode = None

    def __call__(self, argv):

        if self.exitcode is not None:
            # Ignore all but the first call, since if die is called
            # then we certainly want to honor that exitcode, even
            # the ebuild process manages to send a second exit
            # command.
            self.reply_hook = None
        else:
            self.exitcode = int(argv[1])

        # (stdout, stderr, returncode)
        return ("", "", 0)



# Copyright 2012 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

from .AsyncScheduler import AsyncScheduler


class TaskScheduler(AsyncScheduler):

    """
    A simple way to handle scheduling of AbstractPollTask instances. Simply
    pass a task iterator into the constructor and call start(). Use the
    poll, wait, or addExitListener methods to be notified when all of the
    tasks have completed.
    """

    def __init__(self, task_iter, **kwargs):
        AsyncScheduler.__init__(self, **kwargs)
        self._task_iter = task_iter

    def _next_task(self):
        return next(self._task_iter)



# Copyright 2015-2020 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import ctypes
import errno
import re


from portage import _encodings, _unicode_encode
from portage.exception import FileNotFound, PermissionDenied

_compressors = {
    "bzip2": {
        "compress": "${PORTAGE_BZIP2_COMMAND} ${BINPKG_COMPRESS_FLAGS}",
        "decompress": "${PORTAGE_BUNZIP2_COMMAND}",
        "decompress_alt": "${PORTAGE_BZIP2_COMMAND} -d",
        "package": "app-arch/bzip2",
    },
    "gzip": {
        "compress": "gzip ${BINPKG_COMPRESS_FLAGS}",
        "decompress": "gzip -d",
        "package": "app-arch/gzip",
    },
    "lz4": {
        "compress": "lz4 ${BINPKG_COMPRESS_FLAGS}",
        "decompress": "lz4 -d",
        "package": "app-arch/lz4",
    },
    "lzip": {
        "compress": "lzip ${BINPKG_COMPRESS_FLAGS}",
        "decompress": "lzip -d",
        "package": "app-arch/lzip",
    },
    "lzop": {
        "compress": "lzop ${BINPKG_COMPRESS_FLAGS}",
        "decompress": "lzop -d",
        "package": "app-arch/lzop",
    },
    "xz": {
        "compress": "xz -T{JOBS} --memlimit-compress=50% -q ${BINPKG_COMPRESS_FLAGS}",
        "decompress": "xz -T{JOBS} -d",
        "package": "app-arch/xz-utils",
    },
    "zstd": {
        "compress": "zstd -T{JOBS} ${BINPKG_COMPRESS_FLAGS}",
        # If the compression windowLog was larger than the default of 27,
        # then --long=windowLog needs to be passed to the decompressor.
        # Therefore, pass a larger --long=31 value to the decompressor
        # if the current architecture can support it, which is true when
        # sizeof(long) is at least 8 bytes.
        "decompress": "zstd -d"
        + (" --long=31" if ctypes.sizeof(ctypes.c_long) >= 8 else ""),
        "package": "app-arch/zstd",
    },
}

_compression_re = re.compile(
    b"^("
    + b"(?P<bzip2>\x42\x5a\x68\x39)|"
    + b"(?P<gzip>\x1f\x8b)|"
    + b"(?P<lz4>(?:\x04\x22\x4d\x18|\x02\x21\x4c\x18))|"
    + b"(?P<lzip>LZIP)|"
    + b"(?P<lzop>\x89LZO\x00\x0d\x0a\x1a\x0a)|"
    + b"(?P<xz>\xfd\x37\x7a\x58\x5a\x00)|"
    + b"(?P<zstd>([\x22-\x28]\xb5\x2f\xfd)))"
)

_max_compression_re_len = 9


def compression_probe(f):
    """
    Identify the compression type of a file. Returns one of the
    following identifier strings:

            bzip2
            gzip
            lz4
            lzip
            lzop
            xz
            zstd

    @param f: a file path, or file-like object
    @type f: str or file
    @return: a string identifying the compression type, or None if the
            compression type is unrecognized
    @rtype str or None
    """

    open_file = isinstance(f, str)
    if open_file:
        try:
            f = open(
                _unicode_encode(f, encoding=_encodings["fs"], errors="strict"),
                mode="rb",
            )
        except OSError as e:
            if e.errno == PermissionDenied.errno:
                raise PermissionDenied(f)
            elif e.errno in (errno.ENOENT, errno.ESTALE):
                raise FileNotFound(f)
            else:
                raise

    try:
        return _compression_probe_file(f)
    finally:
        if open_file:
            f.close()


def _compression_probe_file(f):

    m = _compression_re.match(f.read(_max_compression_re_len))
    if m is not None:
        for k, v in m.groupdict().items():
            if v is not None:
                return k

    return None



# Copyright 1999-2020 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import collections
import re

import portage
from portage.cache import fs_template
from portage.cache import cache_errors
from portage import os
from portage import _unicode_decode
from portage.util import writemsg
from portage.localization import _


class database(fs_template.FsBased):

    validation_chf = "md5"
    chf_types = ("md5", "mtime")

    autocommits = False
    synchronous = False
    # cache_bytes is used together with page_size (set at sqlite build time)
    # to calculate the number of pages requested, according to the following
    # equation: cache_bytes = page_bytes * page_count
    cache_bytes = 1024 * 1024 * 10

    _connection_info_entry = collections.namedtuple(
        "_connection_info_entry", ("connection", "cursor", "pid")
    )

    def __init__(self, *args, **config):
        super().__init__(*args, **config)
        self._import_sqlite()
        self._allowed_keys = ["_eclasses_"]
        self._allowed_keys.extend(self._known_keys)
        self._allowed_keys.extend("_%s_" % k for k in self.chf_types)
        self._allowed_keys_set = frozenset(self._allowed_keys)
        self._allowed_keys = sorted(self._allowed_keys_set)

        self.location = os.path.join(
            self.location, self.label.lstrip(os.path.sep).rstrip(os.path.sep)
        )

        if not self.readonly and not os.path.exists(self.location):
            self._ensure_dirs()

        config.setdefault("autocommit", self.autocommits)
        config.setdefault("cache_bytes", self.cache_bytes)
        config.setdefault("synchronous", self.synchronous)
        # Set longer timeout for throwing a "database is locked" exception.
        # Default timeout in sqlite3 module is 5.0 seconds.
        config.setdefault("timeout", 15)
        self._config = config
        self._db_connection_info = None

    def _import_sqlite(self):
        # sqlite3 is optional with >=python-2.5
        try:
            import sqlite3 as db_module
        except ImportError as e:
            raise cache_errors.InitializationError(self.__class__, e)

        self._db_module = db_module
        self._db_error = db_module.Error

    def _db_escape_string(self, s):
        """meta escaping, returns quoted string for use in sql statements"""
        if not isinstance(s, str):
            # Avoid potential UnicodeEncodeError in python-2.x by
            # only calling str() when it's absolutely necessary.
            s = str(s)
        return "'%s'" % s.replace("'", "''")

    @property
    def _db_cursor(self):
        if (
            self._db_connection_info is None
            or self._db_connection_info.pid != portage.getpid()
        ):
            self._db_init_connection()
        return self._db_connection_info.cursor

    @property
    def _db_connection(self):
        if (
            self._db_connection_info is None
            or self._db_connection_info.pid != portage.getpid()
        ):
            self._db_init_connection()
        return self._db_connection_info.connection

    def _db_init_connection(self):
        config = self._config
        self._dbpath = self.location + ".sqlite"
        # if os.path.exists(self._dbpath):
        # 	os.unlink(self._dbpath)
        connection_kwargs = {}
        connection_kwargs["timeout"] = config["timeout"]
        try:
            if not self.readonly:
                self._ensure_dirs()
            connection = self._db_module.connect(
                database=_unicode_decode(self._dbpath), **connection_kwargs
            )
            cursor = connection.cursor()
            self._db_connection_info = self._connection_info_entry(
                connection, cursor, portage.getpid()
            )
            self._db_cursor.execute(
                "PRAGMA encoding = %s" % self._db_escape_string("UTF-8")
            )
            if not self.readonly and not self._ensure_access(self._dbpath):
                raise cache_errors.InitializationError(
                    self.__class__, "can't ensure perms on %s" % self._dbpath
                )
            self._db_init_cache_size(config["cache_bytes"])
            self._db_init_synchronous(config["synchronous"])
            self._db_init_structures()
        except self._db_error as e:
            raise cache_errors.InitializationError(self.__class__, e)

    def _db_init_structures(self):
        self._db_table = {}
        self._db_table["packages"] = {}
        mytable = "portage_packages"
        self._db_table["packages"]["table_name"] = mytable
        self._db_table["packages"]["package_id"] = "internal_db_package_id"
        self._db_table["packages"]["package_key"] = "portage_package_key"
        create_statement = []
        create_statement.append("CREATE TABLE")
        create_statement.append(mytable)
        create_statement.append("(")
        table_parameters = []
        table_parameters.append(
            "%s INTEGER PRIMARY KEY AUTOINCREMENT"
            % self._db_table["packages"]["package_id"]
        )
        table_parameters.append("%s TEXT" % self._db_table["packages"]["package_key"])
        for k in self._allowed_keys:
            table_parameters.append("%s TEXT" % k)
        table_parameters.append(
            "UNIQUE(%s)" % self._db_table["packages"]["package_key"]
        )
        create_statement.append(",".join(table_parameters))
        create_statement.append(")")

        self._db_table["packages"]["create"] = " ".join(create_statement)

        cursor = self._db_cursor
        for k, v in self._db_table.items():
            if self._db_table_exists(v["table_name"]):
                create_statement = self._db_table_get_create(v["table_name"])
                table_ok, missing_keys = self._db_validate_create_statement(
                    create_statement
                )
                if table_ok:
                    if missing_keys:
                        for k in sorted(missing_keys):
                            cursor.execute(
                                "ALTER TABLE %s ADD COLUMN %s TEXT"
                                % (self._db_table["packages"]["table_name"], k)
                            )
                else:
                    writemsg(_("sqlite: dropping old table: %s\n") % v["table_name"])
                    cursor.execute("DROP TABLE %s" % v["table_name"])
                    cursor.execute(v["create"])
            else:
                cursor.execute(v["create"])

    def _db_table_exists(self, table_name):
        """return true/false dependant on a tbl existing"""
        cursor = self._db_cursor
        cursor.execute(
            'SELECT name FROM sqlite_master WHERE type="table" AND name=%s'
            % self._db_escape_string(table_name)
        )
        return len(cursor.fetchall()) == 1

    def _db_table_get_create(self, table_name):
        """return true/false dependant on a tbl existing"""
        cursor = self._db_cursor
        cursor.execute(
            "SELECT sql FROM sqlite_master WHERE name=%s"
            % self._db_escape_string(table_name)
        )
        return cursor.fetchall()[0][0]

    def _db_validate_create_statement(self, statement):
        missing_keys = None
        if statement == self._db_table["packages"]["create"]:
            return True, missing_keys

        m = re.match(
            r"^\s*CREATE\s*TABLE\s*%s\s*\(\s*%s\s*INTEGER\s*PRIMARY\s*KEY\s*AUTOINCREMENT\s*,(.*)\)\s*$"
            % (
                self._db_table["packages"]["table_name"],
                self._db_table["packages"]["package_id"],
            ),
            statement,
        )
        if m is None:
            return False, missing_keys

        unique_constraints = {self._db_table["packages"]["package_key"]}
        missing_keys = set(self._allowed_keys)
        unique_re = re.compile(r"^\s*UNIQUE\s*\(\s*(\w*)\s*\)\s*$")
        column_re = re.compile(r"^\s*(\w*)\s*TEXT\s*$")
        for x in m.group(1).split(","):
            m = column_re.match(x)
            if m is not None:
                missing_keys.discard(m.group(1))
                continue
            m = unique_re.match(x)
            if m is not None:
                unique_constraints.discard(m.group(1))
                continue

        if unique_constraints:
            return False, missing_keys

        return True, missing_keys

    def _db_init_cache_size(self, cache_bytes):
        cursor = self._db_cursor
        cursor.execute("PRAGMA page_size")
        page_size = int(cursor.fetchone()[0])
        # number of pages, sqlite default is 2000
        cache_size = cache_bytes // page_size
        cursor.execute("PRAGMA cache_size = %d" % cache_size)
        cursor.execute("PRAGMA cache_size")
        actual_cache_size = int(cursor.fetchone()[0])
        del cursor
        if actual_cache_size != cache_size:
            raise cache_errors.InitializationError(
                self.__class__,
                "actual cache_size = "
                + actual_cache_size
                + " does does not match requested size of "
                + cache_size,
            )

    def _db_init_synchronous(self, synchronous):
        cursor = self._db_cursor
        cursor.execute("PRAGMA synchronous = %d" % synchronous)
        cursor.execute("PRAGMA synchronous")
        actual_synchronous = int(cursor.fetchone()[0])
        del cursor
        if actual_synchronous != synchronous:
            raise cache_errors.InitializationError(
                self.__class__,
                "actual synchronous = "
                + actual_synchronous
                + " does does not match requested value of "
                + synchronous,
            )

    def _getitem(self, cpv):
        cursor = self._db_cursor
        cursor.execute(
            "select * from %s where %s=%s"
            % (
                self._db_table["packages"]["table_name"],
                self._db_table["packages"]["package_key"],
                self._db_escape_string(cpv),
            )
        )
        result = cursor.fetchall()
        if len(result) == 1:
            pass
        elif len(result) == 0:
            raise KeyError(cpv)
        else:
            raise cache_errors.CacheCorruption(cpv, "key is not unique")
        result = result[0]
        d = {}
        allowed_keys_set = self._allowed_keys_set
        for column_index, column_info in enumerate(cursor.description):
            k = column_info[0]
            if k in allowed_keys_set:
                v = result[column_index]
                if v is None:
                    # This happens after a new empty column has been added.
                    v = ""
                d[k] = v

        return d

    def _setitem(self, cpv, values):
        update_statement = []
        update_statement.append(
            "REPLACE INTO %s" % self._db_table["packages"]["table_name"]
        )
        update_statement.append("(")
        update_statement.append(
            ",".join([self._db_table["packages"]["package_key"]] + self._allowed_keys)
        )
        update_statement.append(")")
        update_statement.append("VALUES")
        update_statement.append("(")
        values_parameters = []
        values_parameters.append(self._db_escape_string(cpv))
        for k in self._allowed_keys:
            values_parameters.append(self._db_escape_string(values.get(k, "")))
        update_statement.append(",".join(values_parameters))
        update_statement.append(")")
        cursor = self._db_cursor
        try:
            s = " ".join(update_statement)
            cursor.execute(s)
        except self._db_error as e:
            writemsg("{}: {}\n".format(cpv, str(e)))
            raise

    def commit(self):
        self._db_connection.commit()

    def _delitem(self, cpv):
        cursor = self._db_cursor
        cursor.execute(
            "DELETE FROM %s WHERE %s=%s"
            % (
                self._db_table["packages"]["table_name"],
                self._db_table["packages"]["package_key"],
                self._db_escape_string(cpv),
            )
        )

    def __contains__(self, cpv):
        cursor = self._db_cursor
        cursor.execute(
            " ".join(
                [
                    "SELECT %s FROM %s"
                    % (
                        self._db_table["packages"]["package_id"],
                        self._db_table["packages"]["table_name"],
                    ),
                    "WHERE %s=%s"
                    % (
                        self._db_table["packages"]["package_key"],
                        self._db_escape_string(cpv),
                    ),
                ]
            )
        )
        result = cursor.fetchall()
        if len(result) == 0:
            return False
        if len(result) == 1:
            return True
        raise cache_errors.CacheCorruption(cpv, "key is not unique")

    def __iter__(self):
        """generator for walking the dir struct"""
        cursor = self._db_cursor
        cursor.execute(
            "SELECT %s FROM %s"
            % (
                self._db_table["packages"]["package_key"],
                self._db_table["packages"]["table_name"],
            )
        )
        result = cursor.fetchall()
        key_list = [x[0] for x in result]
        del result
        while key_list:
            yield key_list.pop()
